# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/deduction.py

"""Useful proof manipulation maneuvers in Predicate Logic."""
from .syntax import ForbiddenVariableError, Term, Formula
from .proofs import *
from .prover import *

def remove_assumption(proof: Proof, assumption: Formula,
                      print_as_proof_forms: bool = False) -> Proof:
    """Converts the given proof of some `conclusion` formula, an assumption of
    which is `assumption`, to a proof of '(assumption->conclusion)' from the
    same assumptions except `assumption`.

    Parameters:
        proof: valid proof to convert, from assumptions/axioms that include
            `~predicates.prover.Prover.AXIOMS`.
        assumption: formula that is a simple assumption (i.e., without any
            templates) of the given proof, such that no line of the given proof
            is a UG line over a variable name that is free in this assumption.
        print_as_proof_forms: flag specifying whether the proof of
            '(assumption`->conclusion)' is to be printed in real time as it is
            being created.

    Returns:
        A valid proof of '(assumption->conclusion)'
        from the same assumptions/axioms as the given proof except `assumption`.
    """
    assert proof.is_valid()
    assert Schema(assumption) in proof.assumptions
    assert proof.assumptions.issuperset(Prover.AXIOMS)
    for line in proof.lines:
        if isinstance(line, Proof.UGLine):
            assert line.formula.variable not in assumption.free_variables()

    prover = Prover(proof.assumptions-{Schema(assumption)},print_as_proof_forms)

    for number, line in enumerate(proof.lines):
        if isinstance(line, Proof.AssumptionLine):
            if line.formula == assumption:
                step = prover.add_tautology(Formula('->', assumption, assumption))
            else:
                step = prover.add_instantiated_assumption(line.formula, line.assumption, line.instantiation_map)
                step = prover.add_tautological_implication(Formula('->', assumption, line.formula), [step])

        elif isinstance(line, Proof.TautologyLine):
            step = prover.add_tautology(line.formula)
            step = prover.add_tautology(Formula('->', assumption, line.formula))

        elif isinstance(line, Proof.MPLine):
            antecedent = Formula('->', assumption, proof.lines[line.antecedent_line_number].formula)
            conditional = Formula('->', assumption, proof.lines[line.conditional_line_number].formula)
            consequent = Formula('->', assumption, line.formula)
            antecedent_line_number = None
            conditional_line_number = None
            for line_number, temp_line in enumerate(prover._lines):
                if not antecedent_line_number:
                    if temp_line.formula == antecedent:
                        antecedent_line_number = line_number
                if not conditional_line_number:
                    if temp_line.formula == conditional:
                        conditional_line_number = line_number
                if antecedent_line_number and conditional_line_number:
                    break
            assert isinstance(antecedent_line_number, int) and isinstance(conditional_line_number, int), print('Missing antecedent or conditional. Antecedent_line_number: ', antecedent_line_number, 'conditional_line_number: ', conditional_line_number)
            #consider adding antecedent or conditional if not present?
            step = prover.add_tautological_implication(consequent, {antecedent_line_number, conditional_line_number})

        else:
            assert isinstance(line, Proof.UGLine)
            nonquantified_line_number = None
            nonquantified_formula = proof.lines[line.nonquantified_line_number].formula
            new_nonquantified_formula = Formula('->', assumption, nonquantified_formula)
            new_nonquantified_line_number = None

            for line_number, temp_line in enumerate(prover._lines):
                if temp_line.formula == nonquantified_formula:
                    nonquantified_line_number = line_number
                    break

            if nonquantified_line_number == None:
                for line_number, temp_line in enumerate(prover._lines):
                    if temp_line.formula == new_nonquantified_formula:
                        new_nonquantified_line_number = line_number
                        break
                step1 = prover.add_ug(Formula('A', line.formula.variable, new_nonquantified_formula), new_nonquantified_line_number)
                inst_map = {'Q': assumption, 'R': nonquantified_formula.substitute({line.formula.variable:Term('_')}), 'x': line.formula.variable}
                step2 = prover.add_instantiated_assumption(Prover.US.instantiate(inst_map), Prover.US, inst_map)
                step = prover.add_mp(prover._lines[step2].formula.second, step1, step2)
            else:
                step = prover.add_ug(line.formula, nonquantified_line_number)
                step = prover.add_tautological_implication(Formula('->', assumption, line.formula), {step})
    return prover.qed().clean()
    # Task 11.1

def prove_by_way_of_contradiction(proof: Proof, assumption: Formula) -> Proof:
    """Converts the given proof of a contradiction, an assumption of which is
    `assumption`, to a proof of `'~`assumption`'` from the same
    assumptions except `assumption`.

    Parameters:
        proof: valid proof of a contradiction (i.e., a formula whose negation is
            a tautology) to convert, from assumptions/axioms that include
            `~predicates.prover.Prover.AXIOMS`.
        assumption: formula that is a simple assumption (i.e., without any
            templates) of the given proof, such that no line of the given proof
            is a UG line over a variable name that is free in this assumption.

    Returns:
        A valid proof of `'~`assumption`'` from the same
        assumptions/axioms as the given proof except `assumption`.
    """
    assert proof.is_valid()
    assert Schema(assumption) in proof.assumptions
    assert proof.assumptions.issuperset(Prover.AXIOMS)
    for line in proof.lines:
        if isinstance(line, Proof.UGLine):
            assert line.formula.variable not in assumption.free_variables()
    conditional_proof = remove_assumption(proof, assumption)
    prover = Prover(conditional_proof.assumptions)
    conditional = prover.add_proof(conditional_proof.conclusion, conditional_proof)
    prover.add_tautological_implication(Formula('~', assumption), {conditional})
    return prover.qed()
    # Task 11.2



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/prenex.py

"""Conversion of predicate-logic formulas into prenex normal form."""

from typing import Tuple
from functools import lru_cache
from mltp.logic_utils import fresh_variable_name_generator, is_z_and_number
from .syntax import ForbiddenVariableError, Term, Formula
from .proofs import *
from .prover import *
from .deduction import *

#: Additional axioms of quantification for Predicate Logic.
ADDITIONAL_QUANTIFICATION_AXIOMS = (
    Schema(Formula.parse('((~Ax[R(x)]->Ex[~R(x)])&(Ex[~R(x)]->~Ax[R(x)]))'),
           {'x', 'R'}),
    Schema(Formula.parse('((~Ex[R(x)]->Ax[~R(x)])&(Ax[~R(x)]->~Ex[R(x)]))'),
           {'x', 'R'}),
    Schema(Formula.parse('(((Ax[R(x)]&Q())->Ax[(R(x)&Q())])&'
                         '(Ax[(R(x)&Q())]->(Ax[R(x)]&Q())))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Ex[R(x)]&Q())->Ex[(R(x)&Q())])&'
                         '(Ex[(R(x)&Q())]->(Ex[R(x)]&Q())))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Q()&Ax[R(x)])->Ax[(Q()&R(x))])&'
                         '(Ax[(Q()&R(x))]->(Q()&Ax[R(x)])))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Q()&Ex[R(x)])->Ex[(Q()&R(x))])&'
                         '(Ex[(Q()&R(x))]->(Q()&Ex[R(x)])))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Ax[R(x)]|Q())->Ax[(R(x)|Q())])&'
                         '(Ax[(R(x)|Q())]->(Ax[R(x)]|Q())))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Ex[R(x)]|Q())->Ex[(R(x)|Q())])&'
                         '(Ex[(R(x)|Q())]->(Ex[R(x)]|Q())))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Q()|Ax[R(x)])->Ax[(Q()|R(x))])&'
                         '(Ax[(Q()|R(x))]->(Q()|Ax[R(x)])))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Q()|Ex[R(x)])->Ex[(Q()|R(x))])&'
                         '(Ex[(Q()|R(x))]->(Q()|Ex[R(x)])))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Ax[R(x)]->Q())->Ex[(R(x)->Q())])&'
                         '(Ex[(R(x)->Q())]->(Ax[R(x)]->Q())))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Ex[R(x)]->Q())->Ax[(R(x)->Q())])&'
                         '(Ax[(R(x)->Q())]->(Ex[R(x)]->Q())))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Q()->Ax[R(x)])->Ax[(Q()->R(x))])&'
                         '(Ax[(Q()->R(x))]->(Q()->Ax[R(x)])))'), {'x','R','Q'}),
    Schema(Formula.parse('(((Q()->Ex[R(x)])->Ex[(Q()->R(x))])&'
                         '(Ex[(Q()->R(x))]->(Q()->Ex[R(x)])))'), {'x','R','Q'}),
    Schema(Formula.parse('(((R(x)->Q(x))&(Q(x)->R(x)))->'
                         '((Ax[R(x)]->Ay[Q(y)])&(Ay[Q(y)]->Ax[R(x)])))'),
           {'x', 'y', 'R', 'Q'}),
    Schema(Formula.parse('(((R(x)->Q(x))&(Q(x)->R(x)))->'
                         '((Ex[R(x)]->Ey[Q(y)])&(Ey[Q(y)]->Ex[R(x)])))'),
           {'x', 'y', 'R', 'Q'}))

def is_quantifier_free(formula: Formula) -> bool:
    """Checks if the given formula contains any quantifiers.

    Parameters:
        formula: formula to check.

    Returns:
        `False` if the given formula contains any quantifiers, `True`
        otherwise.
    """
    return formula.quantifiers() == set()
    # Task 11.3a

def is_in_prenex_normal_form(formula: Formula) -> bool:
    """Checks if the given formula is in prenex normal form.

    Parameters:
        formula: formula to check.

    Returns:
        `True` if the given formula in prenex normal form, `False`
        otherwise.
    """
    while is_quantifier(formula.root):
        formula = formula.statement
    return is_quantifier_free(formula)
    # Task 11.3b

def equivalence_of(formula1: Formula, formula2: Formula) -> Formula:
    """States the equivalence of the two given formulas as a formula.

    Parameters:
        formula1: first of the formulas the equivalence of which is to be
            stated.
        formula2: second of the formulas the equivalence of which is to be
            stated.

    Returns:
        The formula `'((`formula1`->`formula2`)&(`formula2`->`formula1`))'`.
    """
    return Formula('&', Formula('->', formula1, formula2),
                   Formula('->', formula2, formula1))
#     return Formula('<->', formula1, formula2)

@lru_cache(maxsize=128)
def has_uniquely_named_variables(formula: Formula) -> bool:
    """Checks if the given formula has uniquely named variables.

    Parameters:
        formula: formula to check.

    Returns:
        `False` if in the given formula some variable name has both bound and
        free occurrences or is quantified by more than one quantifier, `True`
        otherwise.

    Examples:
        >>> has_uniquely_named_variables(
        ...     Formula.parse('(x=0&(Ax[R(x)]|Ex[R(x)]))'))
        False
        >>> has_uniquely_named_variables(
        ...     Formula.parse('(x=0&(Ax[R(x)]|Ey[R(y)]))'))
        False
        >>> has_uniquely_named_variables(
        ...     Formula.parse('(y=0&(Ax[R(x)]|Ex[R(x)]))'))
        False
        >>> has_uniquely_named_variables(
        ...     Formula.parse('(x=0&(Ay[R(y)]|Ez[R(z)]))'))
        True
    """
    forbidden_variables = set(formula.free_variables())
    def has_uniquely_named_variables_helper(formula: Formula) -> bool:
        if is_unary(formula.root):
            return has_uniquely_named_variables_helper(formula.first)
        elif is_binary(formula.root):
            return has_uniquely_named_variables_helper(formula.first) and \
                   has_uniquely_named_variables_helper(formula.second)
        elif is_quantifier(formula.root):
            if formula.variable in forbidden_variables:
                return False
            forbidden_variables.add(formula.variable)
            return has_uniquely_named_variables_helper(formula.statement)
        else:
            assert is_equality(formula.root) or is_relation(formula.root)
            return True

    return has_uniquely_named_variables_helper(formula)

def _uniquely_rename_quantified_variables_helper(formula):
    root = formula.root
    if is_binary(root):
        first, proof1 = _uniquely_rename_quantified_variables_helper(formula.first)
        second, proof2 = _uniquely_rename_quantified_variables_helper(formula.second)
        new_formula = Formula(root, first, second)
        steps = set()
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
        if proof1:
            step1 = prover.add_proof(proof1.conclusion, proof1)
            steps.add(step1)
        if proof2:
            step2 = prover.add_proof(proof2.conclusion, proof2)
            steps.add(step2)
        step3 = prover.add_tautological_implication(equivalence_of(formula, new_formula), steps)

        proof = prover.qed()
        return new_formula, proof

    elif is_unary(root):
        new_formula, proof_of_affirmative = _uniquely_rename_quantified_variables_helper(formula.first)
        new_formula = Formula('~', new_formula)

        proof_of_negative = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
        step1 = proof_of_negative.add_proof(proof_of_affirmative.conclusion, proof_of_affirmative)
        step2 = proof_of_negative.add_tautological_implication(equivalence_of(formula, new_formula), {step1})
        proof = proof_of_negative.qed()
        return new_formula, proof

    elif is_relation(root) or is_equality(root):
        return formula, None

    elif is_quantifier(root):
        if root == 'E':
            schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1]
        else:
            schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2]

        variable = formula.variable
        new_variable = next(fresh_variable_name_generator)
        inner_sub_formula, proof_of_inner_subs = _uniquely_rename_quantified_variables_helper(formula.statement)
        new_formula = Formula(root, new_variable, inner_sub_formula.substitute({variable: Term(new_variable)}))

        if proof_of_inner_subs:
            #then I just need to change one variable to go from inner_sub_formula to new_formula
            prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
            step1 = prover.add_proof(proof_of_inner_subs.conclusion, proof_of_inner_subs)
            inst_map = {
                    'R': formula.statement.substitute({variable:Term('_')}),
                    'Q': inner_sub_formula.substitute({variable: Term('_')}),
                    'x': variable,
                    'y': new_variable}
            step2 = prover.add_instantiated_assumption(schema.instantiate(inst_map), schema, inst_map)
            step3 = prover.add_mp(prover._lines[step2].formula.second, step1, step2)

        else:
            #prove formula -> new_formula
            prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
            inst_map = {'R': formula.statement.substitute({variable: Term('_')}),
                        'Q': formula.statement.substitute({variable: Term('_')}),
                        'x': variable,
                        'y': new_variable}
            step2 = prover.add_instantiated_assumption(schema.instantiate(inst_map), schema, inst_map)
            step3 = prover.add_tautological_implication(prover._lines[step2].formula.second, {step2})
        return new_formula, prover.qed()

def _uniquely_rename_quantified_variables(formula: Formula) -> \
        Tuple[Formula, Proof]:
    """Converts the given formula to an equivalent formula with uniquely named
    variables, and proves the equivalence of these two formulas.

    Parameters:
        formula: formula to convert, which contains no variable names that are
            `z` followed by a number.

    Returns:
        A pair. The first element of the pair is a formula equivalent to the
        given formula, with the exact same structure but with the additional
        property of having uniquely named variables, obtained by consistently
        replacing each variable name that is bound in the given formula with a
        new variable name obtained by calling
        `next`(`~logic_utils.fresh_variable_name_generator`)`. The
        second element of the pair is a proof of the equivalence of the given
        formula and the returned formula (i.e., a proof of
        `equivalence_of`(`formula`,`returned_formula`)`)
        via `~predicates.prover.Prover.AXIOMS` and
        `ADDITIONAL_QUANTIFICATION_AXIOMS`.

    Examples:
        >>> formula = Formula.parse('~(w=x|Aw[(Ex[(x=w&Aw[w=x])]->Ax[x=y])])')
        >>> returned_formula, proof = _uniquely_rename_quantified_variables(
        ...     formula)
        >>> returned_formula
        ~(w=x|Az58[(Ez17[(z17=z58&Az4[z4=z17])]->Az32[z32=y])])
        >>> proof.is_valid()
        True
        >>> proof.conclusion == equivalence_of(formula, returned_formula)
        True
        >>> proof.assumptions == Prover.AXIOMS.union(
        ...     ADDITIONAL_QUANTIFICATION_AXIOMS)
        True
    """
    if isinstance(formula, str):
        formula = Formula.parse(formula)
    for variable in formula.variables():
        assert not is_z_and_number(variable)
    if has_uniquely_named_variables(formula):
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
        prover.add_tautology(equivalence_of(formula, formula))
        return formula, prover.qed()
    else:
        return _uniquely_rename_quantified_variables_helper(formula)
    # Task 11.5

def _pull_out_quantifications_across_negation(formula: Formula) -> \
        Tuple[Formula, Proof]:
    """Converts the given formula with uniquely named variables of the form
    `'~`Q1`x1`[`Q2`x2`[`...`Qn`xn`[`inner_formula`]`...`]]'` to an equivalent
    formula of the form `'`Q'1`x1`[`Q'2`x2`[`...`Q'n`xn`[~`inner_formula`]`...`]]'`,
    and proves the equivalence of these two formulas.

    Parameters:
        formula: formula to convert, whose root is a negation, i.e., which is of
            the form `'~`Q1`x1`[`Q2`x2`[`...`Qn`xn`[`inner_formula`]`...`]]'` where
            `n`>=0, each `Qi` is a quantifier, each `xi` is a variable name, and
            `inner_formula` does not start with a quantifier.

    Returns:
        A pair. The first element of the pair is a formula equivalent to the given
        formula, but of the form `'`Q'1`x1`[`Q'2`x2`[`...`Q'n`xn`[~`inner_formula`]`...`]]'`
        where each `Q'i` is a quantifier, and where the `xi` variable names and `inner_formula`
        are the same as in the given formula. The second element of the pair is a proof of
        the equivalence of the given formula and the returned formula (i.e., a proof of
        `equivalence_of`(`formula`,`returned_formula`)`) via `~predicates.prover.Prover.AXIOMS`
        and `ADDITIONAL_QUANTIFICATION_AXIOMS`.

    Examples:
        >>> formula = Formula.parse('~Ax[Ey[R(x,y)]]')
        >>> returned_formula, proof = _pull_out_quantifications_across_negation(
        ...     formula)
        >>> returned_formula
        Ex[Ay[~R(x,y)]]
        >>> proof.is_valid()
        True
        >>> proof.conclusion == equivalence_of(formula, returned_formula)
        True
        >>> proof.assumptions == Prover.AXIOMS.union(
        ...     ADDITIONAL_QUANTIFICATION_AXIOMS)
        True
    """
    assert is_unary(formula.root)

    if is_quantifier_free(formula):
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
        prover.add_tautology(equivalence_of(formula, formula))
        new_formula = formula
    else:
        variable = formula.first.variable
        statement = formula.first.statement

        if formula.first.root == 'A':
            transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[0]
            quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1] #consider switching this to -1
            new_formula = Formula('E', formula.first.variable, Formula('~', statement))
        else:
            transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[1]
            quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2] #and this to -2
            new_formula = Formula('A', formula.first.variable, Formula('~', statement))

        new_statement, new_statement_proof = _pull_out_quantifications_across_negation(new_formula.statement)
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))

        if is_quantifier_free(statement):
            inst_map = {'R': statement.substitute({variable: Term('_')}), 'x': variable}
            prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)

        else:
            step1 = prover.add_proof(new_statement_proof.conclusion, new_statement_proof)
            inst_map = {'R': Formula('~', statement).substitute({variable: Term('_')}),
                        'Q': new_statement.substitute({variable: Term('_')}),
                        'x': variable,
                        'y': variable}
            step2 = prover.add_instantiated_assumption(quantification_schema.instantiate(inst_map), quantification_schema, inst_map)
            step3 = prover.add_mp(prover._lines[step2].formula.second, step1, step2)
            new_formula = prover._lines[step3].formula.first.second

            inst_map = {'R': statement.substitute({variable: Term('_')}),
                        'x': variable}
            step4 = prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)
            step5 = prover.add_tautological_implication(equivalence_of(formula, new_formula), {step3, step4})
    return new_formula, prover.qed()
    # Task 11.6

def _pull_out_quantifications_from_left_across_binary_operator(formula:
                                                               Formula) -> \
        Tuple[Formula, Proof]:
    """Converts the given formula with uniquely named variables of the form
    `'(`Q1x1`[`Q2x2`[`... `Qnxn`[`inner_first`]`... `]]`*``second`)'`
    to an equivalent formula of the form
    `'`Q'1x1`[`Q'2x2`[`... `Q'nxn`[(`inner_first``*``second`)]`... `]]'`
    and proves the equivalence of these two formulas.

    Parameters:
        formula: formula with uniquely named variables to convert, whose root
            is a binary operator, i.e., which is of the form
            `'(`Q1``x1`[`Q2``x2`[`... `Qn``xn`[`inner_first`]`... `]]`*``second`)'`
            where `*` is a binary operator, `n`>=0, each `Qi` is a quantifier,
            each `xi` is a variable name, and `inner_first` does not start with
            a quantifier.

    Returns:
        A pair. The first element of the pair is a formula equivalent to the
        given formula, but of the form
        `'`Q'1``x1`[`Q'2``x2`[`... `Q'n``xn`[(`inner_first``*``second`)]`... `]]'`
        where each `Q'i` is a quantifier, and where the operator `*`, the `xi`
        variable names, `inner_first`, and `second` are the same as in the given
        formula. The second element of the pair is a proof of the equivalence of
        the given formula and the returned formula (i.e., a proof of
        `equivalence_of`(`formula`,`returned_formula`)`)
        via `~predicates.prover.Prover.AXIOMS` and
        `ADDITIONAL_QUANTIFICATION_AXIOMS`.

    Examples:
        >>> formula = Formula.parse('(Ax[Ey[R(x,y)]]&Ez[P(1,z)])')
        >>> returned_formula, proof = _pull_out_quantifications_from_left_across_binary_operator(
        ...     formula)
        >>> returned_formula
        Ax[Ey[(R(x,y)&Ez[P(1,z)])]]
        >>> proof.is_valid()
        True
        >>> proof.conclusion == equivalence_of(formula, returned_formula)
        True
        >>> proof.assumptions == Prover.AXIOMS.union(
        ...     ADDITIONAL_QUANTIFICATION_AXIOMS)
        True
    """
    root = formula.root
    assert has_uniquely_named_variables(formula)
    assert is_binary(root)

    if not is_quantifier(formula.first.root):
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
        prover.add_tautology(equivalence_of(formula, formula))
        new_formula = formula

    else:
        quantifier = formula.first.root
        variable = formula.first.variable
        first_statement = formula.first.statement
        second = formula.second
        assert formula == Formula(root, Formula(quantifier, variable, first_statement), second)
        new_formula = Formula(quantifier, variable, Formula(root, first_statement, second))
        new_statement = new_formula.statement
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))

        if root == '&':
            if quantifier == 'A':
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[2]
                #(∀x[φ(x)]&ψ)<->∀x[(φ(x)&ψ)]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2]
                #((φ(x)<->ψ(x))->(∀x[φ(x)]<->∀y[ψ(y)]))

            else:
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[3]
                #(∃x[φ(x)]&ψ)<->∃x[(φ(x)&ψ)]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1]
                #((φ(x)<->ψ(x))->(∃x[φ(x)]<->∃y[ψ(y)]))

        elif root == '|':
            if quantifier == 'A':
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[6]
                #(∀x[φ(x)]|ψ)<->∀x[(φ(x)|ψ)]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2]
                ##((φ(x)<->ψ(x))->(∀x[φ(x)]<->∀y[ψ(y)]))

            else:
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[7]
                #(∃x[φ(x)]|ψ)<->∃x[(φ(x)|ψ)]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1]
                #((φ(x)<->ψ(x))->(∃x[φ(x)]<->∃y[ψ(y)])

        elif root == '->':
            if quantifier == 'A':
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[10]
                #(∀x[φ(x)]→ψ) is equivalent to ∃x[(φ(x)→ψ)]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1]
                ##((φ(x)<->ψ(x))->(∀x[φ(x)]<->∀y[ψ(y)]))
                new_formula = Formula('E', variable, Formula(root, first_statement, second))

            else:
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[11]
                #(∃x[φ(x)]→ψ) is equivalent to ∀x[(φ(x)→ψ)]’
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2]
                #((φ(x)<->ψ(x))->(∃x[φ(x)]<->∃y[ψ(y)])
                new_formula = Formula('A', variable, Formula(root, first_statement, second))


        if is_binary(new_formula.statement.root):
            intermediate_formula, intermediate_proof = _pull_out_quantifications_from_left_across_binary_operator(new_formula.statement)
#             step1: prove inner transformation
            step1 = prover.add_proof(intermediate_proof.conclusion, intermediate_proof)
            if is_quantifier_free(intermediate_formula):
                #prove (Ax[T(x)]&S()) <-> Ax[T(x)&S()]
                inst_map = {'R': first_statement.substitute({variable: Term('_')}),
                        'Q': second.substitute({variable: Term('_')}),
                        'x': variable}
                prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)

            else:
#                 step2: prove (Ax[T(x)]&S()) <-> Ax[(T(x)&S())] -> Ay[(Ax[T(x)]&S())] <-> Ay[Ax[(T(x)&S())]] by quantification schema
                inst_map = {'R': new_formula.statement.substitute({variable: Term('_')}),
                            'Q': intermediate_formula.substitute({variable: Term('_')}),
                            'x': variable,
                            'y': variable}
                step2 = prover.add_instantiated_assumption(quantification_schema.instantiate(inst_map), quantification_schema, inst_map)

#                 step3: prove Ay[(Ax[T(x)]&S())] <-> Ay[Ax[(T(x)&S())]] by mp from step1 and step2
                step3 = prover.add_mp(prover._lines[step2].formula.second, step1, step2)

#                 step4: prove (Ay[Ax[T(x)]]&S()) <-> Ay[(Ax[T(x)]&S())] by transfer schema
                inst_map = {'R': first_statement.substitute({variable: Term('_')}),
                            'Q': second.substitute({variable: Term('_')}),
                            'x': variable}
                step4 = prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)

#                 step5: taut imp from step3 and step4 to prove (Ay[Ax[T(x)]]&S()) <-> Ay[Ax[(T(x)&S())]]
                new_formula = prover._lines[step3].formula.first.second
                step5 = prover.add_tautological_implication(equivalence_of(formula, new_formula), {step3, step4})
        else:
            inst_map = {'R': first_statement.substitute({variable: Term('_')}),
                        'Q': second.substitute({variable: Term('_')}),
                        'x': variable}
            prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)
    return new_formula, prover.qed()
    # Task 11.7a

def _pull_out_quantifications_from_right_across_binary_operator(formula:
                                                                Formula) -> \
        Tuple[Formula, Proof]:
    """Converts the given formula with uniquely named variables of the form
    `'(`first`*`Q1`x1`[`Q2`x2`[`...`Qn`xn`[`inner_second`]`...`]])'`
    to an equivalent formula of the form
    `'`Q'1`x1`[`Q'2`x2`[`...`Q'n`xn`[(`first`*`inner_second`)]`...`]]'`
    and proves the equivalence of these two formulas.

    Parameters:
        formula: formula with uniquely named variables to convert, whose root
            is a binary operator, i.e., which is of the form
            `'(`first`*`Q1`x1`[`Q2`x2`[`...`Qn`xn`[`inner_second`]`...`]])'`
            where `*` is a binary operator, `n`>=0, each `Qi` is a quantifier,
            each `xi` is a variable name, and `inner_second` does not start with
            a quantifier.

    Returns:
        A pair. The first element of the pair is a formula equivalent to the
        given formula, but of the form
        `'`Q'1`x1`[`Q'2`x2`[`...`Q'n`xn`[(`first`*`inner_second`)]`...`]]'`
        where each `Q'i` is a quantifier, and where the operator `*`, the `xi`
        variable names, `first`, and `inner_second` are the same as in the given
        formula. The second element of the pair is a proof of the equivalence of
        the given formula and the returned formula (i.e., a proof of
        `equivalence_of`(`formula`,`returned_formula`)`)
        via `~predicates.prover.Prover.AXIOMS` and
        `ADDITIONAL_QUANTIFICATION_AXIOMS`.

    Examples:
        >>> formula = Formula.parse('(Ax[Ey[R(x,y)]]|Ez[P(1,z)])')
        >>> returned_formula, proof = _pull_out_quantifications_from_right_across_binary_operator(
        ...     formula)
        >>> returned_formula
        Ez[(Ax[Ey[R(x,y)]]|P(1,z))]
        >>> proof.is_valid()
        True
        >>> proof.conclusion == equivalence_of(formula, returned_formula)
        True
        >>> proof.assumptions == Prover.AXIOMS.union(
        ...     ADDITIONAL_QUANTIFICATION_AXIOMS)
        True
    """
    root = formula.root
    assert has_uniquely_named_variables(formula)
    assert is_binary(root)

    if not is_quantifier(formula.second.root):
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
        prover.add_tautology(equivalence_of(formula, formula))
        new_formula = formula

    else:
        quantifier = formula.second.root
        variable = formula.second.variable
        second_statement = formula.second.statement
        first = formula.first
        assert formula == Formula(root, first, Formula(quantifier, variable, second_statement))
        new_formula = Formula(quantifier, variable, Formula(root, first, second_statement))
        new_statement = new_formula.statement
        prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))

        if root == '&':
            if quantifier == 'A':
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[4]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2]

            else:
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[5]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1]

        elif root == '|':
            if quantifier == 'A':
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[8]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2]

            else:
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[9]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1]

        elif root == '->':
            if quantifier == 'A':
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[12]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2]

            else:
                transfer_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[13]
                quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-1]

        if is_binary(new_formula.statement.root):
            intermediate_formula, intermediate_proof = _pull_out_quantifications_from_right_across_binary_operator(new_formula.statement)
            step1 = prover.add_proof(intermediate_proof.conclusion, intermediate_proof)

            if is_quantifier_free(intermediate_formula):
                inst_map = {'R': second_statement.substitute({variable: Term('_')}),
                        'Q': first.substitute({variable: Term('_')}),
                        'x': variable}
                prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)

            else:
                inst_map = {'R': new_formula.statement.substitute({variable: Term('_')}),
                            'Q': intermediate_formula.substitute({variable: Term('_')}),
                            'x': variable,
                            'y': variable}
                step2 = prover.add_instantiated_assumption(quantification_schema.instantiate(inst_map), quantification_schema, inst_map)

                step3 = prover.add_mp(prover._lines[step2].formula.second, step1, step2)

                inst_map = {'R': second_statement.substitute({variable: Term('_')}),
                            'Q': first.substitute({variable: Term('_')}),
                            'x': variable}
                step4 = prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)

                new_formula = prover._lines[step3].formula.first.second
                step5 = prover.add_tautological_implication(equivalence_of(formula, new_formula), {step3, step4})
        else:
            inst_map = {'R': second_statement.substitute({variable: Term('_')}),
                        'Q': first.substitute({variable: Term('_')}),
                        'x': variable}
            prover.add_instantiated_assumption(transfer_schema.instantiate(inst_map), transfer_schema, inst_map)
    return new_formula, prover.qed()
    # Task 11.7b

def _pull_out_quantifications_across_binary_operator(formula: Formula) -> \
        Tuple[Formula, Proof]:
    """Converts the given formula with uniquely named variables of the form
    `'(`Q1`x1`[`Q2`x2`[`...`Qn`xn`[`inner_first`]`...`]]`*`P1`y1`[`P2`y2`[`...`Pm`ym`[`inner_second`]`...`]])'`
    to an equivalent formula of the form
    `'`Q'1`x1`[`Q'2`x2`[`...`Q'n`xn`[`P'1`y1`[`P'2`y2`[`...`P'm`ym`[(`inner_first`*`inner_second`)]`...`]]]`...`]]'`
    and proves the equivalence of these two formulas.

    Parameters:
        formula: formula with uniquely named variables to convert, whose root
            is a binary operator, i.e., which is of the form
            `'(`Q1`x1`[`Q2`x2`[`...`Qn`xn`[`inner_first`]`...`]]`*`P1`y1`[`P2`y2`[`...`Pm`ym`[`inner_second`]`...`]])'`
            where `*` is a binary operator, `n`>=0, `m`>=0, each `Qi` and `Pi`
            is a quantifier, each `xi` and `yi` is a variable name, and neither
            `inner_first` nor `inner_second` starts with a quantifier.

    Returns:
        A pair. The first element of the pair is a formula equivalent to the
        given formula, but of the form
        `'`Q'1`x1`[`Q'2`x2`[`...`Q'n`xn`[`P'1`y1`[`P'2`y2`[`...`P'm`ym`[(`inner_first`*`inner_second`)]`...`]]]`...`]]'`
        where each `Q'i` and `P'i` is a quantifier, and where the operator `*`,
        the `xi` and `yi` variable names, `inner_first`, and `inner_second` are
        the same as in the given formula. The second element of the pair is a
        proof of the equivalence of the given formula and the returned formula
        (i.e., a proof of
        `equivalence_of`(`formula`,`returned_formula`)`)
        via `~predicates.prover.Prover.AXIOMS` and
        `ADDITIONAL_QUANTIFICATION_AXIOMS`.

    Examples:
        >>> formula = Formula.parse('(Ax[Ey[R(x,y)]]->Ez[P(1,z)])')
        >>> returned_formula, proof = _pull_out_quantifications_across_binary_operator(
        ...     formula)
        >>> returned_formula
        Ex[Ay[Ez[(R(x,y)->P(1,z))]]]
        >>> proof.is_valid()
        True
        >>> proof.conclusion == equivalence_of(formula, returned_formula)
        True
        >>> proof.assumptions == Prover.AXIOMS.union(
        ...     ADDITIONAL_QUANTIFICATION_AXIOMS)
        True
    """
    assert has_uniquely_named_variables(formula)
    assert is_binary(formula.root)

    formula1, proof1 = _pull_out_quantifications_from_left_across_binary_operator(formula)
    prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
    step1 = prover.add_proof(proof1.conclusion, proof1)

    inner_second = formula1
    quantification_layers = []
    while is_quantifier(inner_second.root):
        quantification_layers.append((inner_second.root, inner_second.variable))
        inner_second = inner_second.statement

    formula2, proof2 = _pull_out_quantifications_from_right_across_binary_operator(inner_second)
    stepx = prover.add_proof(proof2.conclusion, proof2)
    stepz = None
    for quantifier, variable in quantification_layers[::-1]:
        quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2] if quantifier == 'A' else ADDITIONAL_QUANTIFICATION_AXIOMS[-1]
        inst_map = {'R': inner_second.substitute({variable: Term('_')}),
                    'Q': formula2.substitute({variable: Term('_')}),
                    'x': variable,
                    'y': variable}
        stepy = prover.add_instantiated_assumption(quantification_schema.instantiate(inst_map), quantification_schema, inst_map)
        stepz = prover.add_mp(prover._lines[stepy].formula.second, stepx, stepy)
        inner_second = prover._lines[stepz].formula.first.first
        formula2 = prover._lines[stepz].formula.first.second
        stepx = stepz
    if stepz:
        prover.add_tautological_implication(equivalence_of(formula, formula2), {step1, stepz})
    return formula2, prover.qed()
    # Task 11.8

def _to_prenex_normal_form_from_uniquely_named_variables(formula: Formula) -> \
        Tuple[Formula, Proof]:
    """Converts the given formula with uniquely named variables to an equivalent
    formula in prenex normal form, and proves the equivalence of these two
    formulas.

    Parameters:
        formula: formula with uniquely named variables to convert.

    Returns:
        A pair. The first element of the pair is a formula equivalent to the
        given formula, but in prenex normal form. The second element of the pair
        is a proof of the equivalence of the given formula and the returned
        formula (i.e., a proof of
        `equivalence_of`(`formula`,`returned_formula`)`)
        via `~predicates.prover.Prover.AXIOMS` and
        `ADDITIONAL_QUANTIFICATION_AXIOMS`.

    Examples:
        >>> formula = Formula.parse('(~(Ax[Ey[R(x,y)]]->Ez[P(1,z)])|S(w))')
        >>> returned_formula, proof = _to_prenex_normal_form_from_uniquely_named_variables(
        ...     formula)
        >>> returned_formula
        Ax[Ey[Az[(~(R(x,y)->P(1,z))|S(w))]]]
        >>> proof.is_valid()
        True
        >>> proof.conclusion == equivalence_of(formula, returned_formula)
        True
        >>> proof.assumptions == Prover.AXIOMS.union(
        ...     ADDITIONAL_QUANTIFICATION_AXIOMS)
        True
    """
    assert has_uniquely_named_variables(formula)

    prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))
    if is_in_prenex_normal_form(formula):
        new_formula = formula
        prover.add_tautology(equivalence_of(formula, new_formula))

    else:
        root = formula.root

        if is_unary(root):
            new_first, new_first_proof = _to_prenex_normal_form_from_uniquely_named_variables(formula.first)
            new_formula, new_proof = _pull_out_quantifications_across_negation(Formula('~', new_first))

            step1 = prover.add_proof(new_first_proof.conclusion, new_first_proof)
            step2 = prover.add_proof(new_proof.conclusion, new_proof)

            if not new_proof.conclusion == equivalence_of(formula, new_formula):
                step3 = prover.add_tautological_implication(equivalence_of(formula, new_formula), {step1, step2})

        elif is_binary(root):
            new_first, new_first_proof = _to_prenex_normal_form_from_uniquely_named_variables(formula.first)
            new_second, new_second_proof = _to_prenex_normal_form_from_uniquely_named_variables(formula.second)
            new_formula, new_proof = _pull_out_quantifications_across_binary_operator(Formula(root, new_first, new_second))

            step1 = prover.add_proof(new_first_proof.conclusion, new_first_proof)
            step2 = prover.add_proof(new_second_proof.conclusion, new_second_proof)
            step3 = prover.add_tautological_implication(equivalence_of(formula, Formula(root, new_first, new_second)), {step1, step2})
            step4 = prover.add_proof(new_proof.conclusion, new_proof)
            step5 = prover.add_tautological_implication(equivalence_of(formula, new_formula), {step3, step4})

        elif is_quantifier(root):
            variable = formula.variable
            new_statement, new_statement_proof = _to_prenex_normal_form_from_uniquely_named_variables(formula.statement)
            new_formula = Formula(root, formula.variable, new_statement)

            step1 = prover.add_proof(new_statement_proof.conclusion, new_statement_proof)
            #use quantification schema to build back up?
            quantification_schema = ADDITIONAL_QUANTIFICATION_AXIOMS[-2] if root == 'A' else ADDITIONAL_QUANTIFICATION_AXIOMS[-1]
            inst_map = {'R': formula.statement.substitute({variable: Term('_')}),
                        'Q': new_statement.substitute({variable: Term('_')}),
                        'x': variable,
                        'y': variable}
            step2 = prover.add_instantiated_assumption(quantification_schema.instantiate(inst_map), quantification_schema, inst_map)
            step3 = prover.add_tautological_implication(prover._lines[step2].formula.second, {step1, step2})

    return new_formula, prover.qed()
    # Task 11.9

def to_prenex_normal_form(formula: Formula) -> Tuple[Formula, Proof]:
    """Converts the given formula to an equivalent formula in prenex normal
    form, and proves the equivalence of these two formulas.

    Parameters:
        formula: formula to convert, which contains no variable names that are
            `z` followed by a number.

    Returns:
        A pair. The first element of the pair is a formula equivalent to the
        given formula, but in prenex normal form. The second element of the pair
        is a proof of the equivalence of the given formula and the returned
        formula (i.e., a proof of
        `equivalence_of`(`formula`,`returned_formula`)`)
        via `~predicates.prover.Prover.AXIOMS` and
        `ADDITIONAL_QUANTIFICATION_AXIOMS`.

    Examples:
        >>> formula = Formula.parse('~(w=x|Aw[(Ex[(x=w&Aw[w=x])]->Ax[x=y])])')
        >>> returned_formula, proof = to_prenex_normal_form(formula)
        >>> returned_formula
        Ez58[Ez17[Az4[Ez32[~(w=x|((z17=z58&z4=z17)->z32=y))]]]]
        >>> proof.is_valid()
        True
        >>> proof.conclusion == equivalence_of(formula, returned_formula)
        True
        >>> proof.assumptions == Prover.AXIOMS.union(
        ...     ADDITIONAL_QUANTIFICATION_AXIOMS)
        True
    """
    for variable in formula.variables():
        assert not is_z_and_number(variable)

    prover = Prover(Prover.AXIOMS.union(ADDITIONAL_QUANTIFICATION_AXIOMS))

    if is_in_prenex_normal_form(formula):
        prover.add_tautology(equivalence_of(formula, formula))
        prenex_formula = formula

    elif has_uniquely_named_variables(formula):
        prenex_formula, prenex_proof = _to_prenex_normal_form_from_uniquely_named_variables(formula)
        prover.add_proof(prenex_proof.conclusion, prenex_proof)

    else:
        renamed_formula, renaming_proof = _uniquely_rename_quantified_variables(formula)
        prenex_formula, prenex_proof = _to_prenex_normal_form_from_uniquely_named_variables(renamed_formula)

        step1 = prover.add_proof(renaming_proof.conclusion, renaming_proof)
        step2 = prover.add_proof(prenex_proof.conclusion, prenex_proof)
        step3 = prover.add_tautological_implication(equivalence_of(formula, prenex_formula), {step1, step2})

    return prenex_formula, prover.qed().clean()
    # Task 11.10



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/functions.py

"""Syntactic conversion of predicate-logic formulas to not use functions and
equality."""

from typing import AbstractSet, List, Set, Union
from collections import Counter
from mltp.logic_utils import fresh_variable_name_generator, is_z_and_number
from functools import reduce
from .syntax import ForbiddenVariableError, Term, Formula
from .semantics import *

def function_name_to_relation_name(function: str) -> str:
    """Converts the given function name to a canonically corresponding relation
    name.

    Parameters:
        function: function name to convert.

    Returns:
        A relation name that is the same as the given function name, except that
        its first letter is capitalized.
    """
    assert is_function(function)
    return function[0].upper() + function[1:]

def relation_name_to_function_name(relation: str) -> str:
    """Converts the given relation name to a canonically corresponding function
    name.

    Parameters:
        relation: relation name to convert.

    Returns:
        A function name `function` such that
        `function_name_to_relation_name`\ ``(``\ `function`\ ``)`` is the given
        relation name.
    """
    assert is_relation(relation)
    return relation[0].lower() + relation[1:]

def replace_functions_with_relations_in_model(model: Model[T]) -> Model[T]:
    """Converts the given model to a canonically corresponding model without any
    function interpretations, replacing each function interpretation with a
    canonically corresponding relation interpretation.

    Parameters:
        model: model to convert, such that there exist no canonically
            corresponding function name and relation name that both have
            interpretations in this model.

    Returns:
        A model obtained from the given model by replacing every function
        interpretation of a function name with a relation interpretation of the
        canonically corresponding relation name, such that the relation
        interpretation contains any tuple
        ``(``\ `x1`\ ``,``...\ ``,``\ `xn`\ ``)``  if and only if `x1` is the
        output of the function interpretation for the arguments
        ``(``\ `x2`\ ``,``...\ ``,``\ `xn`\ ``)``.
    """
    for function in model.function_interpretations:
        assert function_name_to_relation_name(function) not in \
               model.relation_interpretations

    new_relation_interpretations = dict(model.relation_interpretations)

    for function_name in model.function_interpretations:
        new_relation_interpretations[function_name_to_relation_name(function_name)] = {(model.function_interpretations[function_name][arguments],*arguments) for arguments in model.function_interpretations[function_name]}

    return Model(model.universe, model.constant_interpretations, new_relation_interpretations, dict())
    # Task 8.1

def replace_relations_with_functions_in_model(model: Model[T],
                                              original_functions:
                                              AbstractSet[str]) -> \
        Union[Model[T], None]:
    """Converts the given model with no function interpretations to a
    canonically corresponding model with interpretations for the given function
    names, having each new function interpretation replace a canonically
    corresponding relation interpretation.

    Parameters:
        model: model to convert, which contains no function interpretations.
        original_functions: function names for the model to convert to,
            such that no relation name that canonically corresponds to any of
            these function names has an interpretation in the given model.

    Returns:
        A model `model` with the given function names such that
        `replace_functions_with_relations_in_model`\ ``(``\ `model`\ ``)``
        is the given model, or ``None`` if no such model exists.
    """
    assert len(model.function_interpretations) == 0
    for function in original_functions:
        assert is_function(function)
        assert function not in model.function_interpretations
        assert function_name_to_relation_name(function) in \
               model.relation_interpretations

    new_function_interpretations = dict()
    relation_interpretations = dict(model.relation_interpretations)

    for function_name, relation_name in zip(original_functions, map(function_name_to_relation_name, original_functions)):
        relata = relation_interpretations.pop(relation_name)
        function_dict = {ntuple[1:]:ntuple[0] for ntuple in relata}

        if len(function_dict) != len(relata) or len(function_dict) != len(model.universe)**len(list(function_dict)[0]):
            return None
        else:
            new_function_interpretations[function_name] = function_dict
    return Model(model.universe, model.constant_interpretations, relation_interpretations, new_function_interpretations)
    # Task 8.2

def _compile_term(term: Term) -> List[Formula]:
    """Syntactically compiles the given term into a list of single-function
    invocation steps.

    Parameters:
        term: term to compile, whose root is a function invocation, and which
            contains no variable names that are ``z`` followed by a number.

    Returns:
        A list of steps, each of which is a formula of the form 'y = f(x1,...,xn)',
        where `y` is a new variable name obtained by calling
        `next(logic_utils.fresh_variable_name_generator)`, `f` is a function name,
        and each of the `xi` is either a constant name or a variable name. If `xi`
        is a new variable name, then it is also the left-hand side of a previous
        step, where all of the steps "leading up to" `x1` precede those "leading
        up" to `x2`, etc. If all the returned steps hold in any model, then the
        left-hand-side variable name of the last returned step evaluates in that
        model to the value of the given term.
    """
    assert is_function(term.root)
    for variable in term.variables():
        assert not is_z_and_number(variable)

    pile = []
    new_arguments = []
    for argument in term.arguments:
        if is_function(argument.root):
            for i in _compile_term(argument):
                pile.append(i)
            new_arguments.append(pile[-1].arguments[0])
        else:
            new_arguments.append(argument)
    pile.append(Formula('=',[Term.parse(next(fresh_variable_name_generator)),Term(term.root, new_arguments)]))
    return pile
    # Task 8.3

def replace_functions_with_relations_in_formula(formula: Formula) -> Formula:
    """Syntactically converts the given formula to a formula that does not
    contain any function invocations, and is "one-way equivalent" in the sense
    that the former holds in a model if and only if the latter holds in the
    canonically corresponding model with no function interpretations.

    Parameters:
        formula: formula to convert, which contains no variable names that are
            ``z`` followed by a number, and such that there exist no canonically
            corresponding function name and relation name that are both invoked
            in this formula.

    Returns:
        A formula such that the given formula holds in any model `model` if and
        only if the returned formula holds in
        `replace_function_with_relations_in_model(model)`.
    """
    assert len({function_name_to_relation_name(function) for
                function,arity in formula.functions()}.intersection(
                    {relation for relation,arity in formula.relations()})) == 0
    for variable in formula.variables():
        assert not is_z_and_number(variable)

    root = formula.root

    if is_relation(root) or is_equality(root):
        pile = []
        new_arguments = []
        for argument in formula.arguments:
            if is_function(argument.root):
                pile.append(_compile_term(argument))
                new_arguments.append(pile[-1][-1].arguments[0])
            else:
                new_arguments.append(argument)
        new_formula = Formula(formula.root, new_arguments)

        for compiled_formula in pile[::-1]:
            for step in compiled_formula[::-1]:
                new_relation = Formula(function_name_to_relation_name(step.arguments[1].root),
                                        [step.arguments[0], *step.arguments[1].arguments])
                new_formula = Formula('E', str(step.arguments[0]),
                                        Formula('&', new_relation, new_formula))

    elif is_binary(root):
        new_formula = Formula(formula.root,
                                replace_functions_with_relations_in_formula(formula.first),
                                replace_functions_with_relations_in_formula(formula.second))
    elif is_unary(root):
        new_formula = Formula(formula.root,
                                replace_functions_with_relations_in_formula(formula.first))
    else:
        new_formula = Formula(formula.root,
                                formula.variable,
                                replace_functions_with_relations_in_formula(formula.statement))
    return new_formula
    # Task 8.4

def replace_functions_with_relations_in_formulas(formulas:
                                                 AbstractSet[Formula]) -> \
        Set[Formula]:
    """Syntactically converts the given set of formulas to a set of formulas
    that do not contain any function invocations, and is "two-way
    equivalent" in the sense that:

    1. The former holds in a model if and only if the latter holds in the
       canonically corresponding model with no function interpretations.
    2. The latter holds in a model if and only if that model has a
       canonically corresponding model with interpretations for the functions
       names of the former, and the former holds in that model.

    Parameters:
        formulas: formulas to convert, which contain no variable names that are
            ``z`` followed by a number, and such that there exist no canonically
            corresponding function name and relation name that are both invoked
            in these formulas.

    Returns:
        A set of formulas, one for each given formula as well as one additional
        formula for each relation name that replaces a function name from the
        given formulas, such that:

        1. The given formulas hold in a model `model` if and only if the
           returned formulas hold in
           `replace_functions_with_relations_in_model(model)`.
        2. The returned formulas hold in a model `model` if and only if
           `replace_relations_with_functions_in_model(model, original_functions)`,
           where `original_functions` are all the function names in the given
           formulas, is a model and the given formulas hold in it.
    """
    assert len(set.union(*[{function_name_to_relation_name(function) for
                            function,arity in formula.functions()}
                           for formula in formulas]).intersection(
                               set.union(*[{relation for relation,arity in
                                            formula.relations()} for
                                           formula in formulas]))) == 0
    for formula in formulas:
        for variable in formula.variables():
            assert not is_z_and_number(variable)

    new_formulas = {replace_functions_with_relations_in_formula(formula) for formula in formulas}
    new_relations = set.union(*[i.relations() for i in new_formulas]) - set.union(*[i.relations() for i in formulas])

    for relation in new_relations:
        arguments = [Term.parse(next(fresh_variable_name_generator)) for _ in range(relation[1]+2)]

        existence_formula = Formula('E', str(arguments[0]),
                                Formula(relation[0], [arguments[0], *arguments[3:]]))
        uniqueness_formula = Formula('->',
                                Formula('&',
                                    Formula(relation[0], [arguments[1], *arguments[3:]]),
                                    Formula(relation[0], [arguments[2], *arguments[3:]])),
                                Formula('=',
                                    [arguments[1],
                                    arguments[2]]))

        for argument in arguments[1:]:
            existence_formula = Formula('A', str(argument), existence_formula)
            uniqueness_formula = Formula('A', str(argument), uniqueness_formula)

        new_formulas.add(Formula('&', existence_formula, uniqueness_formula))
    return new_formulas
    # Task 8.5

def translate_equality_to_SAME(formula):
    root = formula.root
    if root == '=':
        formula = Formula('SAME', formula.arguments)
    else:
        if is_unary(root):
            formula = Formula(root, translate_equality_to_SAME(formula.first))

        elif is_binary(root):
            formula = Formula(root,
                        translate_equality_to_SAME(formula.first),
                        translate_equality_to_SAME(formula.second))

        elif is_quantifier(root):
            formula = Formula(root, formula.variable, translate_equality_to_SAME(formula.statement))
    return formula

    #helper function for task 8.6

def replace_equality_with_SAME_in_formulas(formulas: AbstractSet[Formula]) -> \
        Set[Formula]:
    """Syntactically converts the given set of formulas to a canonically
    corresponding set of formulas that do not contain any equalities, consisting
    of the following formulas:

    1. A formula for each of the given formulas, where each equality is
       replaced with a matching invocation of the relation name ``'SAME'``.
    2. Formula(s) that ensure that in any model of the returned formulas, the
       interpretation of the relation name ``'SAME'`` is reflexive,
       symmetric, and transitive.
    3. For each relation name from the given formulas, formula(s) that ensure
       that in any model of the returned formulas, the interpretation of this
       relation name respects the interpretation of the relation name
       ``'SAME'``.

    Parameters:
        formulas: formulas to convert, that contain no function names and do not
            contain the relation name ``'SAME'``.

    Returns:
        The converted set of formulas.
    """
    for formula in formulas:
        assert len(formula.functions()) == 0
        assert 'SAME' not in \
               {relation for relation,arity in formula.relations()}

    translations = [translate_equality_to_SAME(formula) for formula in formulas]

    new_formulas = [*translations,
                    Formula.parse('Ax[SAME(x,x)]'),                                  # reflexivity
                    Formula.parse('Ax[Ay[(SAME(x,y)<->SAME(y,x))]]'),                # symmetry
                    Formula.parse('Ax[Ay[Az[((SAME(x,y)&SAME(y,z))->SAME(x,z))]]]')] # transitivity

#   For every relation:
    relations = set.union(*[i.relations() for i in formulas])

#   1. I need two sets of variables, each as long as the relation arity
    variables1, variables2 = [], []
    max_arity = max(i[1] for i in relations)
    for i in range(max_arity*2):
        if i%2:
            variables2.append(Term.parse(next(fresh_variable_name_generator)))
        else:
            variables1.append(Term.parse(next(fresh_variable_name_generator)))

    for relation in set.union(*[i.relations() for i in formulas]):
        varset1 = variables1[:relation[1]]
        varset2 = variables2[:relation[1]]

#       2. Express R(*var1) -> R(*var2)
        relation_formula = Formula('->',
                            Formula(relation[0], varset1),
                            Formula(relation[0], varset2))

#       3. Pair variables and build SAME formulae for each pair
        same_formula = None
        for pair in zip(varset1, varset2):
            if same_formula:
                same_formula = Formula('&', same_formula, Formula('SAME', [pair[0],pair[1]]))
            else:
                same_formula = Formula('SAME', [pair[0],pair[1]])

#       4. new formula = SAME formulae -> equivalence formula
        new_formula = Formula('->', same_formula, relation_formula)

#       5. Universally quantify over each of those variables to get Ax1[...Axn[Ay1[...Ayn[((SAME(x1,y1)&...&SAME(xn,yn))->(R(x1,...xn)->R(y1,...yn)))]]]]
        for var in varset1+varset2:
            new_formula = Formula('A', str(var), new_formula)
        new_formulas.append(new_formula)

    return new_formulas
    # Task 8.6

def add_SAME_as_equality_in_model(model: Model[T]) -> Model[T]:
    """Adds an interpretation of the relation name 'SAME' in the given
    model, that canonically corresponds to equality in the given model.

    Parameters:
        model: model that has no interpretation of the relation name
            ``'SAME'``, to add the interpretation to.

    Returns:
        A model obtained from the given model by adding an interpretation of the
        relation name 'SAME', that contains precisely all pairs
        `(x,x)` for every element `x` of the universe of
        the given model.
    """
    assert 'SAME' not in model.relation_interpretations
    new_relation_interpretations = dict(model.relation_interpretations)
    new_relation_interpretations['SAME'] = {(i,i) for i in model.universe}
    return Model(model.universe, model.constant_interpretations, new_relation_interpretations, model.function_interpretations)
    # Task 8.7

def make_equality_as_SAME_in_model(model: Model[T]) -> Model[T]:
    """Converts the given model to a model where equality coincides with the
    interpretation of ``'SAME'`` in the given model, in the sense that any set
    of formulas holds in the returned model if and only if its canonically
    corresponding set of formulas that do not contain equality holds in the
    given model.

    Parameters:
        model: model to convert, that contains no function interpretations, and
            contains an interpretation of the relation name ``'SAME'`` that is
            reflexive, symmetric, transitive, and respected by the
            interpretations of all other relation names.

    Returns:
        A model that is a model of any set `formulas` if and only if the given
        model is a model of
        `replace_equality_with_SAME`\ ``(``\ `formulas`\ ``)``. The universe of
        the returned model corresponds to the equivalence classes of the
        interpretation of ``'SAME'`` in the given model.
    """
    assert 'SAME' in model.relation_interpretations and \
           model.relation_arities['SAME'] == 2
    assert len(model.function_interpretations) == 0

    equivalence_classes = []
    same = [set(i) for i in model.relation_interpretations['SAME'] if len(i) > 1]

    #find equivalence classes
    for element in model.universe:
        temp_eq_class = set.union(*[double for double in same if element in double])
        if equivalence_classes:
            for equivalence_class in equivalence_classes:
                if any([member in equivalence_class for member in temp_eq_class]):
                    equivalence_class = equivalence_class | temp_eq_class
                    break
                else:
                    equivalence_classes.append(temp_eq_class)
        else:
            equivalence_classes.append(temp_eq_class)

    #make a dictionary and new_universe
    equivalence_dict = dict()
    new_universe = set()
    for _class in equivalence_classes:
        value = list(_class)[0]
        new_universe.add(value)
        for key in _class:
            equivalence_dict[key] = value

    new_constant_interpretations = dict()
    for constant in model.constant_interpretations:
        new_constant_interpretations[constant] = equivalence_dict[model.constant_interpretations[constant]]

    new_relation_interpretations = dict()#{'SAME': {(i,i) for i in new_universe}}
    for relation in model.relation_interpretations:
        if relation == 'SAME':
            continue
        new_tuples = []
        for _tuple in model.relation_interpretations[relation]:
            new_tuples.append(tuple(equivalence_dict[i] for i in _tuple))
        new_relation_interpretations[relation] = new_tuples
    return Model(new_universe, new_constant_interpretations, new_relation_interpretations, model.function_interpretations)
    # Task 8.8

# term = Term.parse('f(g(x,h(0)),f(f(0,g(y)),h(h(x))))')
# print(term,'\n')
# pile = _compile_term(term)

# formula = Formula.parse('R(f(g(x)),h(2,y),3)')
#print(replace_functions_with_relations_in_formula(formula))
# model = Model({'0', '1', '2'}, {'a': '0', 'b': '1', 'c': '2'},
#                   {'SAME': {('0','0'), ('1','1'), ('2','2'), ('1','2'),
#                             ('2','1')},
#                    'Q': {('0',), ('1',), ('2',)}})
# new_model = make_equality_as_SAME_in_model(model)



# mltp/logic/propositions/__init__.py
from .syntax import (is_variable,
                     is_constant,
                     is_unary,
                     is_binary,
                     split_str,
                     Formula)
from .semantics import (Model,
                        is_model,
                        variables,
                        evaluate,
                        all_models,
                        truth_values,
                        print_truth_table,
                        is_tautology,
                        is_contradiction,
                        is_satisfiable,
                        _synthesize_for_model,
                        synthesize,
                        _synthesize_for_all_except_model,
                        synthesize_cnf,
                        combine_formulae,
                        evaluate_inference,
                        is_sound_inference)
from .operators import (to_not_and_or,
                        to_not_and,
                        to_nand,
                        to_implies_not,
                        to_implies_false,
                        to_not_or,
                        to_nor)
from .reductions import (Graph,
                         is_graph,
                         is_valid_3coloring,
                         graph3coloring_to_formula,
                         assignment_to_3coloring,
                         tricolor_graph)
from .axiomatic_systems import (MP,
                                I0,
                                I1,
                                D,
                                I2,
                                N,
                                NI,
                                NN,
                                R,
                                AXIOMATIC_SYSTEM,
                                HILBERT_AXIOMATIC_SYSTEM,
                                A,
                                NA1,
                                NA2,
                                O1,
                                O2,
                                NO,
                                T,
                                NF,
                                AXIOMATIC_SYSTEM_FULL,
                                N_ALTERNATIVE,
                                HILBERT_AXIOMATIC_SYSTEM_ALTERNATIVE,
                                AE1,
                                AE2,
                                OE,
                                HILBERT_AXIOMATIC_SYSTEM_FULL)
from .proofs import (SpecializationMap,
                     InferenceRule,
                     Proof,
                     prove_specialization,
                     _inline_proof_once,
                     uses_rule,
                     find_first_use_of_rule,
                     inline_proof)
from .deduction import (prove_corollary,
                        combine_proofs,
                        remove_assumption,
                        prove_from_opposites,
                        prove_by_way_of_contradiction)
from .soundness import (rule_nonsoundness_from_specialization_nonsoundness,
                        nonsound_rule_of_nonsound_proof)
from .tautology import (formulas_capturing_model,
                        prove_in_model,
                        reduce_assumption,
                        prove_tautology,
                        proof_or_counterexample,
                        encode_as_formula,
                        prove_sound_inference,
                        model_or_inconsistency,
                        prove_in_model_full)

#__all__ = ['deduction', 'operators', 'proofs', 'reductions', 'semantics', 'some_proofs', 'soundness', 'syntax', 'tautology']

__all__ = ['is_variable',
             'is_constant',
             'is_unary',
             'is_binary',
             'split_str',
             'Formula',
             'Model',
             'is_model',
             'variables',
             'evaluate',
             'all_models',
             'truth_values',
             'print_truth_table',
             'is_tautology',
             'is_contradiction',
             'is_satisfiable',
             '_synthesize_for_model',
             'synthesize',
             '_synthesize_for_all_except_model',
             'synthesize_cnf',
             'combine_formulae',
             'evaluate_inference',
             'is_sound_inference',
             'to_not_and_or',
             'to_not_and',
             'to_nand',
             'to_implies_not',
             'to_implies_false',
             'to_not_or',
             'to_nor',
             'Graph',
             'is_graph',
             'is_valid_3coloring',
             'graph3coloring_to_formula',
             'assignment_to_3coloring',
             'tricolor_graph',
             'MP',
             'I0',
             'I1',
             'D',
             'I2',
             'N',
             'NI',
             'NN',
             'R',
             'AXIOMATIC_SYSTEM',
             'HILBERT_AXIOMATIC_SYSTEM',
             'A',
             'NA1',
             'NA2',
             'O1',
             'O2',
             'NO',
             'T',
             'NF',
             'AXIOMATIC_SYSTEM_FULL',
             'N_ALTERNATIVE',
             'HILBERT_AXIOMATIC_SYSTEM_ALTERNATIVE',
             'AE1',
             'AE2',
             'OE',
             'HILBERT_AXIOMATIC_SYSTEM_FULL',
             'SpecializationMap',
             'InferenceRule',
             'Proof',
             'prove_specialization',
             '_inline_proof_once',
             'uses_rule',
             'find_first_use_of_rule',
             'inline_proof',
             'prove_corollary',
             'combine_proofs',
             'remove_assumption',
             'prove_from_opposites',
             'prove_by_way_of_contradiction', 
             'rule_nonsoundness_from_specialization_nonsoundness',
             'nonsound_rule_of_nonsound_proof',
             'formulas_capturing_model',
             'prove_in_model',
             'reduce_assumption',
             'prove_tautology',
             'proof_or_counterexample',
             'encode_as_formula',
             'prove_sound_inference',
             'model_or_inconsistency',
             'prove_in_model_full']


# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/some_proofs.py

"""Some proofs in Predicate Logic."""
from .syntax import ForbiddenVariableError, Term, Formula
from .proofs import *
from .prover import *
from .deduction import *
from .prenex import equivalence_of

def prove_syllogism(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the assumptions:

    1. All men are mortal (`'Ax[(Man(x)->Mortal(x))]'`), and
    2. Aristotle is a man (`'Man(aristotle)'`)

    the conclusion: Aristotle is mortal (`'Mortal(aristotle)'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'Ax[(Man(x)->Mortal(x))]', 'Man(aristotle)'},
                    print_as_proof_forms)
    step1 = prover.add_assumption('Ax[(Man(x)->Mortal(x))]')
    step2 = prover.add_instantiated_assumption(
        '(Ax[(Man(x)->Mortal(x))]->(Man(aristotle)->Mortal(aristotle)))',
        Prover.UI, {'R': '(Man(_)->Mortal(_))', 'c': 'aristotle'})
    step3 = prover.add_mp('(Man(aristotle)->Mortal(aristotle))', step1, step2)
    step4 = prover.add_assumption('Man(aristotle)')
    step5 = prover.add_mp('Mortal(aristotle)', step4, step3)
    return prover.qed()

def prove_syllogism_with_universal_instantiation(print_as_proof_forms: bool =
                                                 False) -> Proof:
    """Using the method `~predicates.prover.Prover.add_universal_instantiation`,
    proves from the assumptions:

    1. All men are mortal (`'Ax[(Man(x)->Mortal(x))]'`), and
    2. Aristotle is a man (`'Man(aristotle)'`)

    the conclusion: Aristotle is mortal (`'Mortal(aristotle)'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof, created with the help of the method
        `~predicates.prover.Prover.add_universal_instantiation`, of the above
        inference via `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'Ax[(Man(x)->Mortal(x))]', 'Man(aristotle)'},
                    print_as_proof_forms)
    step1 = prover.add_assumption('Ax[(Man(x)->Mortal(x))]')
    step2 = prover.add_universal_instantiation('(Man(aristotle)->Mortal(aristotle))', step1, 'aristotle')
    step3 = prover.add_assumption('Man(aristotle)')
    step4 = prover.add_mp('Mortal(aristotle)', step3, step2)
    return prover.qed()

def prove_syllogism_all_all(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the assumptions:

    1. All Greeks are human (`'Ax[(Greek(x)->Human(x))]'`), and
    2. All humans are mortal (`'Ax[(Human(x)->Mortal(x))]'`)

    the conclusion: All Greeks are mortal (`'Ax[(Greek(x)->Mortal(x))]'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'Ax[(Greek(x)->Human(x))]', 'Ax[(Human(x)->Mortal(x))]'},
                    print_as_proof_forms)
    step1 = prover.add_assumption('Ax[(Greek(x)->Human(x))]')
    step2 = prover.add_universal_instantiation(
        '(Greek(x)->Human(x))', step1, 'x')
    step3 = prover.add_assumption('Ax[(Human(x)->Mortal(x))]')
    step4 = prover.add_universal_instantiation(
        '(Human(x)->Mortal(x))', step3, 'x')
    step5 = prover.add_tautology(
        '((Greek(x)->Human(x))->((Human(x)->Mortal(x))->(Greek(x)->Mortal(x))))')
    step6 = prover.add_mp(
        '((Human(x)->Mortal(x))->(Greek(x)->Mortal(x)))', step2, step5)
    step7 = prover.add_mp('(Greek(x)->Mortal(x))', step4, step6)
    step8 = prover.add_ug('Ax[(Greek(x)->Mortal(x))]', step7)
    return prover.qed()

def prove_syllogism_all_all_with_tautological_implication(print_as_proof_forms:
                                                          bool = False) -> \
        Proof:
    """Using the method
    `~predicates.prover.Prover.add_tautological_implication`, proves from the
    assumptions:

    1. All Greeks are human (`'Ax[(Greek(x)->Human(x))]'`), and
    2. All humans are mortal (`'Ax[(Human(x)->Mortal(x))]'`)

    the conclusion: All Greeks are mortal (`'Ax[(Greek(x)->Mortal(x))]'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof, created with the help of the method
        `~predicates.prover.Prover.add_tautological_implication`, of the above
        inference via `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'Ax[(Greek(x)->Human(x))]', 'Ax[(Human(x)->Mortal(x))]'},
                    print_as_proof_forms)
    step1 = prover.add_assumption('Ax[(Greek(x)->Human(x))]')
    step2 = prover.add_universal_instantiation(
        '(Greek(x)->Human(x))', step1, 'x')
    step3 = prover.add_assumption('Ax[(Human(x)->Mortal(x))]')
    step4 = prover.add_universal_instantiation(
        '(Human(x)->Mortal(x))', step3, 'x')
    step5 = prover.add_tautological_implication(
        '(Greek(x)->Mortal(x))', {step2, step4})
    step6 = prover.add_ug('Ax[(Greek(x)->Mortal(x))]', step5)
    return prover.qed()

def prove_syllogism_all_exists(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the assumptions:

    1. All men are mortal (`'Ax[(Man(x)->Mortal(x))]'`), and
    2. Some men exist (`'Ex[Man(x)]'`)

    the conclusion: Some mortals exist (`'Ex[Mortal(x)]'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'Ax[(Man(x)->Mortal(x))]', 'Ex[Man(x)]'},
                    print_as_proof_forms)
    step1 = prover.add_assumption('Ax[(Man(x)->Mortal(x))]')
    step2 = prover.add_assumption('Ex[Man(x)]')
    step3 = prover.add_universal_instantiation(
        '(Man(x)->Mortal(x))', step1, 'x')
    step4 = prover.add_instantiated_assumption(
        '(Mortal(x)->Ex[Mortal(x)])', Prover.EI,
        {'R': 'Mortal(_)', 'c': 'x'})
    step5 = prover.add_tautological_implication(
        '(Man(x)->Ex[Mortal(x)])', {step3, step4})
    step6 = prover.add_ug('Ax[(Man(x)->Ex[Mortal(x)])]', step5)
    step7 = prover.add_instantiated_assumption(
        '((Ax[(Man(x)->Ex[Mortal(x)])]&Ex[Man(x)])->Ex[Mortal(x)])', Prover.ES,
        {'R': 'Man(_)', 'Q': 'Ex[Mortal(x)]'})
    step8 = prover.add_tautological_implication(
        'Ex[Mortal(x)]', {step2, step6, step7})
    return prover.qed()

def prove_syllogism_all_exists_with_existential_derivation(print_as_proof_forms:
                                                           bool = False) -> \
        Proof:
    """Using the method `~predicates.prover.Prover.add_existential_derivation`,
    proves from the assumptions:

    1. All men are mortal (`'Ax[(Man(x)->Mortal(x))]'`), and
    2. Some men exist (`'Ex[Man(x)]'`)

    the conclusion: Some mortals exist (`'Ex[Mortal(x)]'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof, created with the help of the method
        `~predicates.prover.Prover.add_existential_derivation`, of the above
        inference via `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'Ax[(Man(x)->Mortal(x))]', 'Ex[Man(x)]'},
                    print_as_proof_forms)
    step1 = prover.add_assumption('Ax[(Man(x)->Mortal(x))]')
    step2 = prover.add_assumption('Ex[Man(x)]')
    step3 = prover.add_universal_instantiation(
        '(Man(x)->Mortal(x))', step1, 'x')
    step4 = prover.add_instantiated_assumption(
        '(Mortal(x)->Ex[Mortal(x)])', Prover.EI, {'R': 'Mortal(_)', 'c': 'x'})
    step5 = prover.add_tautological_implication(
        '(Man(x)->Ex[Mortal(x)])', {step3, step4})
    step6 = prover.add_existential_derivation('Ex[Mortal(x)]', step2, step5)
    return prover.qed()

def prove_lovers(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the assumptions:

    1. Everybody loves somebody (`'Ax[Ey[Loves(x,y)]]'`), and
    2. Everybody loves a lover (`'Ax[Az[Ay[(Loves(x,y)->Loves(z,x))]]]'`)

    the conclusion: Everybody loves everybody (`'Ax[Az[Loves(z,x)]]'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being constructed.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'Ax[Ey[Loves(x,y)]]',
                     'Ax[Az[Ay[(Loves(x,y)->Loves(z,x))]]]'},
                    print_as_proof_forms)
    step1 = prover.add_assumption('Ax[Ey[Loves(x,y)]]')
    step2 = prover.add_universal_instantiation('Ey[Loves(x,y)]', step1, 'x')
    step3 = prover.add_assumption('Ax[Az[Ay[(Loves(x,y)->Loves(z,x))]]]')
    step4 = prover.add_universal_instantiation('Az[Ay[(Loves(x,y)->Loves(z,x))]]', step3, 'x')
    step5 = prover.add_universal_instantiation('Ay[(Loves(x,y)->Loves(z,x))]', step4, 'z')
    instantiation_map = {'R':Formula.parse('Loves(x,_)'), 'Q':Formula.parse('Loves(z,x)'), 'x':'y'}
    step6 = prover.add_instantiated_assumption(Prover.ES.instantiate(instantiation_map), Prover.ES, instantiation_map)
    step7 = prover.add_and_introduction(7, step2)
    step8 = prover.add_mp('Loves(z,x)',11,8)
    step9 = prover.add_ug('Az[Loves(z,x)]', step8)
    step10 = prover.add_ug('Ax[Az[Loves(z,x)]]', step9)
    return prover.qed().clean()
    # Task 10.4

def prove_homework(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the assumptions:

    1. No homework is fun (`'~Ex[(Homework(x)&Fun(x))]'`), and
    2. Some reading is homework (`'Ex[(Homework(x)&Reading(x))]'`)

    the conclusion: Some reading is not fun (`'Ex[(Reading(x)&~Fun(x))]'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being constructed.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({'~Ex[(Homework(x)&Fun(x))]',
                     'Ex[(Homework(x)&Reading(x))]'}, print_as_proof_forms)
    step1 = prover.add_assumption('~Ex[(Homework(x)&Fun(x))]')
    step2 = prover.add_instantiated_assumption('((Homework(x)&Fun(x))->Ex[(Homework(x)&Fun(x))])', Prover.EI, {'R': '(Homework(_)&Fun(_))', 'c': 'x'})
    step3 = prover.add_tautological_implication('~(Homework(x)&Fun(x))',{step1,step2})
    step5 = prover.add_assumption('Ex[(Homework(x)&Reading(x))]')
    step6 = prover.add_tautology('(~(Homework(x)&Fun(x))->((Homework(x)&Reading(x))->(Reading(x)&~Fun(x))))')
    step7 = prover.add_mp('((Homework(x)&Reading(x))->(Reading(x)&~Fun(x)))', step3, step6)
    instantiation_map = {'R': Formula.parse('(Reading(_)&~Fun(_))'), 'c': Term('x'), 'x': 'x'}
    step8 = prover.add_instantiated_assumption('((Reading(x)&~Fun(x))->Ex[(Reading(x)&~Fun(x))])', Prover.EI, instantiation_map)
    step9 = prover.add_tautology('(((Homework(x)&Reading(x))->(Reading(x)&~Fun(x)))->(((Reading(x)&~Fun(x))->Ex[(Reading(x)&~Fun(x))])->((Homework(x)&Reading(x))->Ex[(Reading(x)&~Fun(x))])))')
    step10 = prover.add_mp('(((Reading(x)&~Fun(x))->Ex[(Reading(x)&~Fun(x))])->((Homework(x)&Reading(x))->Ex[(Reading(x)&~Fun(x))]))', step7, step9)
    step11 = prover.add_mp('((Homework(x)&Reading(x))->Ex[(Reading(x)&~Fun(x))])', step8, step10)
    step12 = prover.add_existential_derivation('Ex[(Reading(x)&~Fun(x))]', step5, step11)
    return prover.qed().clean()
    # Task 10.5

#: The three group axioms
GROUP_AXIOMS = frozenset({'plus(0,x)=x', 'plus(minus(x),x)=0',
                          'plus(plus(x,y),z)=plus(x,plus(y,z))'})

def prove_group_right_neutral(stop_before_flipped_equality: bool,
                              stop_before_free_instantiation: bool,
                              stop_before_substituted_equality: bool,
                              stop_before_chained_equality: bool,
                              print_as_proof_forms: bool = False) -> Proof:
    """Proves from the group axioms that x+0=x (`'plus(x,0)=x'`).

    Parameters:
        stop_before_flipped_equality: flag specifying to stop proving just
            before the first call to
            `~predicates.prover.Prover.add_flipped_equality` and to return the
            prefix of the proof created up to that point.
        stop_before_free_instantiation: flag specifying to stop proving just
            before the first call to
            `~predicates.prover.Prover.add_free_instantiation` and to return the
            prefix of the proof created up to that point.
        stop_before_substituted_equality: flag specifying stop proving just
            before the first call to
            `~predicates.prover.Prover.add_substituted_equality` and to return
            the prefix of the proof created up to that point.
        stop_before_chained_equality: flag specifying to stop proving just
            before the first call to
            `~predicates.prover.Prover.add_chained_equality` and to return the
            prefix of the proof created up to that point.
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover(GROUP_AXIOMS, print_as_proof_forms)
    zero = prover.add_assumption('plus(0,x)=x')
    negation = prover.add_assumption('plus(minus(x),x)=0')
    associativity = prover.add_assumption('plus(plus(x,y),z)=plus(x,plus(y,z))')
    if stop_before_flipped_equality:
        return prover.qed()
    flipped_zero = prover.add_flipped_equality('x=plus(0,x)', zero)
    flipped_negation = prover.add_flipped_equality(
        '0=plus(minus(x),x)', negation)
    flipped_associativity = prover.add_flipped_equality(
        'plus(x,plus(y,z))=plus(plus(x,y),z)', associativity)
    if stop_before_free_instantiation:
        return prover.qed()
    step7 = prover.add_free_instantiation(
        '0=plus(minus(minus(x)),minus(x))', flipped_negation, {'x': 'minus(x)'})
    step8 = prover.add_flipped_equality(
        'plus(minus(minus(x)),minus(x))=0', step7)
    step9 = prover.add_free_instantiation(
        'plus(plus(minus(minus(x)),minus(x)),x)='
        'plus(minus(minus(x)),plus(minus(x),x))',
        associativity, {'x': 'minus(minus(x))', 'y': 'minus(x)', 'z': 'x'})
    step10 = prover.add_free_instantiation('plus(0,0)=0', zero, {'x': '0'})
    step11 = prover.add_free_instantiation(
        'plus(x,0)=plus(0,plus(x,0))', flipped_zero, {'x': 'plus(x,0)'})
    step12 = prover.add_free_instantiation(
        'plus(0,plus(x,0))=plus(plus(0,x),0)', flipped_associativity,
        {'x': '0', 'y': 'x', 'z': '0'})
    if stop_before_substituted_equality:
        return prover.qed()
    step13 = prover.add_substituted_equality(
        'plus(plus(0,x),0)=plus(plus(plus(minus(minus(x)),minus(x)),x),0)',
        step7, 'plus(plus(_,x),0)')
    step14 = prover.add_substituted_equality(
        'plus(plus(plus(minus(minus(x)),minus(x)),x),0)='
        'plus(plus(minus(minus(x)),plus(minus(x),x)),0)',
        step9, 'plus(_,0)')
    step15 = prover.add_substituted_equality(
        'plus(plus(minus(minus(x)),plus(minus(x),x)),0)='
        'plus(plus(minus(minus(x)),0),0)',
        negation, 'plus(plus(minus(minus(x)),_),0)')
    step16 = prover.add_free_instantiation(
        'plus(plus(minus(minus(x)),0),0)=plus(minus(minus(x)),plus(0,0))',
        associativity, {'x': 'minus(minus(x))', 'y': '0', 'z': '0'})
    step17 = prover.add_substituted_equality(
        'plus(minus(minus(x)),plus(0,0))=plus(minus(minus(x)),0)',
        step10, 'plus(minus(minus(x)),_)')
    step18 = prover.add_substituted_equality(
        'plus(minus(minus(x)),0)=plus(minus(minus(x)),plus(minus(x),x))',
        flipped_negation, 'plus(minus(minus(x)),_)')
    step19 = prover.add_free_instantiation(
        'plus(minus(minus(x)),plus(minus(x),x))='
        'plus(plus(minus(minus(x)),minus(x)),x)', flipped_associativity,
        {'x': 'minus(minus(x))','y': 'minus(x)','z': 'x'})
    step20 = prover.add_substituted_equality(
        'plus(plus(minus(minus(x)),minus(x)),x)=plus(0,x)', step8, 'plus(_,x)')
    if stop_before_chained_equality:
        return prover.qed()
    step21 = prover.add_chained_equality(
        'plus(x,0)=x',
        [step11, step12, step13, step14, step15, step16, step17, step18, step19,
         step20, zero])
    return prover.qed()

def prove_group_unique_zero(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the group axioms and from the assumption a+c=a
    (`'plus(a,c)=a'`) that c=0 (`'c=0'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover(GROUP_AXIOMS.union({'plus(a,c)=a'}), print_as_proof_forms)
    step1 = prover.add_assumption('plus(a,c)=a')
    step2 = prover.add_assumption('plus(0,x)=x')
    step3 = prover.add_ug('Ax[plus(0,x)=x]', step2)
    step4 = prover.add_universal_instantiation('plus(0,a)=a', step3, 'a')
    step5 = prover.add_flipped_equality('a=plus(0,a)', step4)
    step6 = prover.add_chained_equality('plus(a,c)=plus(0,a)', [step1, step5])
    step7 = prover.add_substituted_equality(Formula.parse('plus(minus(a),plus(a,c))=plus(minus(a),plus(0,a))'), step6, Term.parse('plus(minus(a),_)'))
    step8 = prover.add_assumption('plus(plus(x,y),z)=plus(x,plus(y,z))')

    inst_map = {'x': Term.parse('minus(a)'), 'y': 'a', 'z': 'c'}
    step9 = prover.add_free_instantiation('plus(plus(minus(a),a),c)=plus(minus(a),plus(a,c))', step8, inst_map)

    step10 = prover.add_assumption('plus(minus(x),x)=0')
    step11 = prover.add_free_instantiation('plus(minus(a),a)=0', step10, {'x':'a'})
    step12 = prover.add_assumption('plus(0,x)=x')

    inst_map = {'R': Formula.parse('plus(plus(minus(a),a),c)=plus(_,c)'), 'c': Term.parse('plus(minus(a),a)'), 'd': Term('0')}
    step13 = prover.add_instantiated_assumption(Prover.ME.instantiate(inst_map), Prover.ME, inst_map)
    step14 = prover.add_mp('(plus(plus(minus(a),a),c)=plus(plus(minus(a),a),c)->plus(plus(minus(a),a),c)=plus(0,c))', step11, step13)
    step15 = prover.add_instantiated_assumption('plus(plus(minus(a),a),c)=plus(plus(minus(a),a),c)', Prover.RX, {'c': Term.parse('plus(plus(minus(a),a),c)')})
    step16 = prover.add_mp('plus(plus(minus(a),a),c)=plus(0,c)', step15, step14)
    step17 = prover.add_free_instantiation('plus(0,c)=c', step12, {'x':'c'})

    inst_map = {'R': Formula.parse('plus(minus(a),plus(0,a))=plus(minus(a),_)'), 'c': Term.parse('plus(0,a)'), 'd': Term('a')}
    step18 = prover.add_instantiated_assumption(Prover.ME.instantiate(inst_map), Prover.ME, inst_map)
    step19 = prover.add_mp('(plus(minus(a),plus(0,a))=plus(minus(a),plus(0,a))->plus(minus(a),plus(0,a))=plus(minus(a),a))', step4, step18)
    step20 = prover.add_instantiated_assumption(Prover.RX.instantiate({'c':Term.parse('plus(minus(a),plus(0,a))')}), Prover.RX, {'c':'plus(minus(a),plus(0,a))'})
    step21 = prover.add_mp('plus(minus(a),plus(0,a))=plus(minus(a),a)', step20, step19)
    step22 = prover.add_flipped_equality('plus(0,c)=plus(plus(minus(a),a),c)', step16)
    step23 = prover.add_flipped_equality('c=plus(0,c)', step17)
    step24 = prover.add_chained_equality('c=0', [step23, step22, step9, step7, step21, step11])
    return prover.qed().clean()
    # Task 10.10

#: The six field axioms
FIELD_AXIOMS = frozenset(GROUP_AXIOMS.union(
    {'plus(x,y)=plus(y,x)', 'times(x,1)=x', 'times(x,y)=times(y,x)',
     'times(times(x,y),z)=times(x,times(y,z))', '(~x=0->Ey[times(y,x)=1])',
     'times(x,plus(y,z))=plus(times(x,y),times(x,z))'}))

def prove_field_zero_multiplication(print_as_proof_forms: bool = False) -> \
        Proof:
    """Proves from the field axioms that 0*x=0 (`'times(0,x)=0'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover(FIELD_AXIOMS, print_as_proof_forms)
    # add assumptions
    step1_1 = prover.add_assumption('plus(0,x)=x')
    step1_2 = prover.add_assumption('times(x,plus(y,z))=plus(times(x,y),times(x,z))')
    step1_3 = prover.add_assumption('plus(minus(x),x)=0')
    step1_4 = prover.add_assumption('plus(x,y)=plus(y,x)')
    step1_5 = prover.add_assumption('plus(plus(x,y),z)=plus(x,plus(y,z))')
    step1_6 = prover.add_assumption('times(x,y)=times(y,x)')

    # 0+0=0 ; plus(0,0)=0 - free instantiation of axiom 3
    step2 = prover.add_free_instantiation('plus(0,0)=0', step1_1, {'x':'0'})

    # x*0=x*(0+0) ; times(x,0)=times(x,plus(0,0)) - ME
    step3_1 = prover.add_flipped_equality('0=plus(0,0)', step2)
    inst_map = {'R': Formula.parse('times(x,0)=times(x,_)'), 'c': Term.parse('0'), 'd': Term.parse('plus(0,0)')}
    step3_2 = prover.add_instantiated_assumption(Prover.ME.instantiate(inst_map), Prover.ME, inst_map)
    step3_3 = prover.add_mp(prover._lines[step3_2].formula.second, step3_1, step3_2)
    step3_4 = prover.add_instantiated_assumption('times(x,0)=times(x,0)', Prover.RX, {'c': 'times(x,0)'})
    step3_5 = prover.add_mp(prover._lines[step3_3].formula.second, step3_4, step3_3)

    # x*(0+0)=(x*0)+(x*0) ; times(x,plus(0,0))=plus(times(x,0),times(x,0)) - free instantiation of axiom 6
    step4 = prover.add_free_instantiation('times(x,plus(0,0))=plus(times(x,0),times(x,0))', step1_2, {'y':'0', 'z':'0'})

    # x*0=(x*0)+(x*0) ; times(x,0)=plus(times(x,0),times(x,0)) - chain from the previous two
    step5_1 = prover.add_chained_equality('times(x,0)=plus(times(x,0),times(x,0))', [step3_5, step4])

    # -(x*0)+((x*0)+(x*0))=-(x*0)+(x*0) ; ME with c=(x*0)+(x*0), d=x*0, R='-(x*0)+_=-(x*0)+(x*0)'
    inst_map = {'R': Formula.parse('plus(minus(times(x,0)),_)=plus(minus(times(x,0)),times(x,0))'), 'c': Term.parse('times(x,0)'), 'd': Term.parse('plus(times(x,0),times(x,0))')}
    step6_1 = prover.add_instantiated_assumption(Prover.ME.instantiate(inst_map), Prover.ME, inst_map)
    step6_2 = prover.add_mp(prover._lines[step6_1].formula.second, step5_1, step6_1)
    step6_3 = prover.add_instantiated_assumption('plus(minus(times(x,0)),times(x,0))=plus(minus(times(x,0)),times(x,0))', Prover.RX, {'c': 'plus(minus(times(x,0)),times(x,0))'})
    step6_4 = prover.add_mp(prover._lines[step6_2].formula.second, step6_3, step6_2)

    # -(x*0)+(x*0)=0 ; free instantiation of axiom 2
    step7 = prover.add_free_instantiation('plus(minus(times(x,0)),times(x,0))=0', step1_3, {'x':'times(x,0)'})
    step7_1 = prover.add_flipped_equality('0=plus(minus(times(x,0)),times(x,0))', step7)

    # (-(x*0)+(x*0))+(x*0)=0
    step8 = prover.add_chained_equality('plus(minus(times(x,0)),plus(times(x,0),times(x,0)))=0' , [step6_4,step7])

    # 0+(x*0)=(-(x*0)+(x*0))+(x*0)
    inst_map = {'R': Formula.parse('plus(0,times(x,0))=plus(_,times(x,0))'), 'c': Term.parse('0'), 'd': Term.parse('plus(minus(times(x,0)),times(x,0))')}
    step9_1 = prover.add_instantiated_assumption(Prover.ME.instantiate(inst_map), Prover.ME, inst_map)
    step9_2 = prover.add_mp(prover._lines[step9_1].formula.second, step7_1, step9_1)
    step9_3 = prover.add_instantiated_assumption('plus(0,times(x,0))=plus(0,times(x,0))', Prover.RX, {'c': 'plus(0,times(x,0))'})
    step9_4 = prover.add_mp(prover._lines[step9_2].formula.second, step9_3, step9_2)

    # x=plus(x,0) from x=plus(0,x)
    step10_1 = prover.add_flipped_equality('x=plus(0,x)', step1_1)
    step10_2 = prover.add_free_instantiation('plus(0,x)=plus(x,0)', step1_4, {'x': '0', 'y': 'x'})
    step10_3 = prover.add_chained_equality('x=plus(x,0)', [step10_1, step10_2])

    # x*0=(x*0)+0 ; times(x,0)=plus(times(x,0),0)
    step11_1 = prover.add_free_instantiation('times(x,0)=plus(times(x,0),0)', step10_3, {'x': Term.parse('times(x,0)')})

    step12 = prover.add_free_instantiation('plus(times(x,0),0)=plus(0,times(x,0))', step1_4, {'x': 'times(x,0)', 'y': '0'})

    step13 = prover.add_free_instantiation('plus(plus(minus(times(x,0)),times(x,0)),times(x,0))=plus(minus(times(x,0)),plus(times(x,0),times(x,0)))', step1_5, {'x': 'minus(times(x,0))', 'y': 'times(x,0)', 'z': 'times(x,0)'})

    # times(0,x) = times(x,0)
    step14 = prover.add_free_instantiation('times(0,x)=times(x,0)', step1_6, {'x': '0', 'y': 'x'})

    prover.add_chained_equality('times(0,x)=0', [step14, step11_1, step12, step9_4, step13, step8])
    return prover.qed().clean()
    # Task 10.11

#: Axiom schema of induction
INDUCTION_AXIOM = Schema(
    Formula.parse('((R(0)&Ax[(R(x)->R(s(x)))])->Ax[R(x)])'), {'R'})
#: The seven axioms of Peano arithmetic
PEANO_AXIOMS = frozenset({'(s(x)=s(y)->x=y)', '~s(x)=0', 'plus(x,0)=x',
                          'plus(x,s(y))=s(plus(x,y))', 'times(x,0)=0',
                          'times(x,s(y))=plus(times(x,y),x)', INDUCTION_AXIOM})

def prove_peano_left_neutral(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the axioms of Peano arithmetic that 0+x=x
    (`'plus(0,x)=x'`).

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    from mltp.propositions import Formula as PropositionalFormula
    prover = Prover(PEANO_AXIOMS, print_as_proof_forms)

    # prove (plus(0,x)=x->s(plus(0,x))=s(x))
    inst_map = {'R': Formula.parse('s(plus(0,x))=s(_)'), 'c': Term.parse('plus(0,x)'), 'd': Term.parse('x')}
    step1 = prover.add_instantiated_assumption(Prover.ME.instantiate(inst_map), Prover.ME, inst_map)
    step2 = prover.add_instantiated_assumption('s(plus(0,x))=s(plus(0,x))', Prover.RX, {'c':'s(plus(0,x))'})
    taut = Formula.from_propositional_skeleton(PropositionalFormula.parse('(q->(p->q))'), {'p':Formula.parse('plus(0,x)=x'), 'q':Formula.parse('s(plus(0,x))=s(plus(0,x))')})
    step3 = prover.add_tautology(taut)
    step4 = prover.add_mp(prover._lines[step3].formula.second, step2, step3)
    taut = Formula.from_propositional_skeleton(PropositionalFormula.parse('((p->(q->r))->((p->q)->(p->r)))'), {'p':Formula.parse('plus(0,x)=x'), 'q':Formula.parse('s(plus(0,x))=s(plus(0,x))'), 'r':Formula.parse('s(plus(0,x))=s(x)')})
    step5 = prover.add_tautology(taut)
    step6 = prover.add_mp(prover._lines[step5].formula.second, step1, step5)
    step7 = prover.add_mp(prover._lines[step6].formula.second, step4, step6)

    # prove (plus(0,x)=x->plus(0,s(x))=s(x))
    step8 = prover.add_assumption('plus(x,s(y))=s(plus(x,y))')
    step9 = prover.add_free_instantiation('plus(0,s(x))=s(plus(0,x))', step8, {'x': '0', 'y': 'x'})         #this or step10 will be q for the next tautology
    step10 = prover.add_flipped_equality('s(plus(0,x))=plus(0,s(x))', step9)

    inst_map = {'R': Formula.parse('plus(0,s(x))=_'), 'c': Term.parse('s(plus(0,x))'), 'd': Term.parse('s(x)')}
    step10 = prover.add_instantiated_assumption(Prover.ME.instantiate(inst_map), Prover.ME, inst_map)

    taut = PropositionalFormula.parse('((p->(q->r))->(q->(p->r)))')
    inst_map = {'p': Formula.parse('s(plus(0,x))=s(x)'), 'q': Formula.parse('plus(0,s(x))=s(plus(0,x))'), 'r': Formula.parse('plus(0,s(x))=s(x)')}
    step11 = prover.add_tautology(Formula.from_propositional_skeleton(taut,inst_map))
    step12 = prover.add_mp(prover._lines[step11].formula.second, step10, step11)
    step13 = prover.add_mp(prover._lines[step12].formula.second, step9, step12)

    taut= PropositionalFormula.parse('((p->q)->((q->r)->(p->r)))')
    inst_map = {'p': Formula.parse('plus(0,x)=x'), 'q': Formula.parse('s(plus(0,x))=s(x)'), 'r': Formula.parse('plus(0,s(x))=s(x)')}
    step14 = prover.add_tautology(Formula.from_propositional_skeleton(taut, inst_map))
    step15 = prover.add_mp(prover._lines[step14].formula.second, step7, step14)
    step16 = prover.add_mp(prover._lines[step15].formula.second, step13, step15)

    #end game
    step17 = prover.add_assumption('plus(x,0)=x')
    step18 = prover.add_free_instantiation('plus(0,0)=0', step17, {'x':'0'})
    step19 = prover.add_ug(Formula('A','x', prover._lines[step16].formula), step16)
    step20 = prover.add_and_introduction(step18, step19)
    inst_map = {'R': Formula.parse('plus(0,_)=_')}
    step21 = prover.add_instantiated_assumption(INDUCTION_AXIOM.instantiate(inst_map), INDUCTION_AXIOM, inst_map)
    step22 = prover.add_mp('Ax[plus(0,x)=x]', step20, step21)
    step23 = prover.add_universal_instantiation('plus(0,x)=x', step22, 'x')
    return prover.qed().clean()
    # Task 10.12

#: Axiom schema of (unrestricted) comprehension
COMPREHENSION_AXIOM = Schema(
    Formula.parse('Ey[Ax[((In(x,y)->R(x))&(R(x)->In(x,y)))]]'), {'R'})

def prove_russell_paradox(print_as_proof_forms: bool = False) -> Proof:
    """Proves from the axioms schema of unrestricted comprehension the
    contradiction `'(z=z&~z=z)'`.

    Parameters:
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above inference via
        `~predicates.prover.Prover.AXIOMS`.
    """
    prover = Prover({COMPREHENSION_AXIOM}, print_as_proof_forms)
    COMPREHENSION_AXIOM.instantiate({'R': Formula.parse('~In(_,_)')})
    step1 = prover.add_instantiated_assumption(COMPREHENSION_AXIOM.instantiate({'R': Formula.parse('~In(_,_)')}), COMPREHENSION_AXIOM, {'R': Formula.parse('~In(_,_)')})
    inst_map = {'R': Formula.parse('((In(_,y)->~In(_,_))&(~In(_,_)->In(_,y)))'),
                'c': Term.parse('y'),
                'x': 'x'}
    step2 = prover.add_instantiated_assumption(Prover.UI.instantiate(inst_map), Prover.UI, inst_map)
    step3 = prover.add_tautology('(((In(y,y)->~In(y,y))&(~In(y,y)->In(y,y)))->(P()&~P()))')
    step4 = prover.add_tautological_implication('(Ax[((In(x,y)->~In(x,x))&(~In(x,x)->In(x,y)))]->((z=z&~z=z)&(Hegel()->Sux(balls))))', {step2, step3})
    step5 = prover.add_existential_derivation('((z=z&~z=z)&(Hegel()->Sux(balls)))', step1, step4)
    step6 = prover.add_tautological_implication('(Hegel()->Sux(balls))', {step5})
    step7 = prover.add_tautological_implication('(z=z&~z=z)', {step5})
    return prover.qed()
    # Task 10.13

def _prove_not_exists_not_implies_all(variable: str, formula: Formula,
                                      print_as_proof_forms: bool = False) -> \
        Proof:
    """Proves that
    '(~E`variable`[~`formula`]->A`variable`[`formula`])'.

    Parameters:
        variable: variable name for the quantifications in the formula to be
            proven.
        formula: statement to be universally quantified, and whose negation is
            to be existentially quantified, in the formula to be proven.
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above formula via
        `~predicates.prover.Prover.AXIOMS`.
    """
    assert is_variable(variable)
    if isinstance(formula, str):
        formula = Formula.parse(formula)
    not_exists_not_schema = Schema(Formula('~',Formula('E',variable,Formula('~',formula))))

    prover = Prover({not_exists_not_schema},False)
    inst_map = {'R': Formula('~',formula.substitute({variable: Term('_')})), 'c': Term(variable), 'x': variable}
    step1 = prover.add_instantiated_assumption(Prover.EI.instantiate(inst_map), Prover.EI, inst_map)
    step2 = prover.add_assumption(not_exists_not_schema.formula)
    step3 = prover.add_tautological_implication(formula, {step1, step2})
    step4 = prover.add_ug(Formula('A', variable, formula), step3)
    return remove_assumption(prover.qed(), not_exists_not_schema.formula, print_as_proof_forms)
    # Optional Task 11.4a

def _prove_exists_not_implies_not_all(variable: str, formula: Formula,
                                      print_as_proof_forms: bool = False) -> \
        Proof:
    """Proves that
    `'(E`variable`[~`formula`]->~A`variable`[`formula`])'`.

    Parameters:
        variable: variable name for the quantifications in the formula to be
            proven.
        formula: statement to be universally quantified, and whose negation is
            to be existentially quantified, in the formula to be proven.
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above formula via
        `~predicates.prover.Prover.AXIOMS`.
    """
    assert is_variable(variable)
    if isinstance(formula, str):
        formula = Formula.parse(formula)
    exists_not_schema = Schema(Formula('E', variable, Formula('~', formula)))

    prover = Prover({exists_not_schema}, False)
    step1 = prover.add_assumption(exists_not_schema.formula)
    inst_map = {'R': formula.substitute({variable: Term('_')}), 'c': Term(variable), 'x': variable}
    step2 = prover.add_instantiated_assumption(Prover.UI.instantiate(inst_map), Prover.UI, inst_map)
    step3 = prover.add_tautological_implication(Formula('->', Formula('~', formula), Formula('~', Formula('A', variable, formula))), {step2})
    step4 = prover.add_ug(Formula('A', variable, prover._lines[step3].formula), step3)
    inst_map = {'Q': Formula('~', Formula('A', variable, formula)), 'R': Formula('~', formula).substitute({variable: Term('_')}), 'x': variable}
    step5 = prover.add_instantiated_assumption(Prover.ES.instantiate(inst_map), Prover.ES, inst_map)
    step6 = prover.add_tautological_implication(Formula('~', Formula('A', variable, formula)), {step1, step4, step5})
    return remove_assumption(prover.qed(), exists_not_schema.formula, print_as_proof_forms)
    # Optional Task 11.4b

def prove_not_all_iff_exists_not(variable: str, formula: Formula,
                                 print_as_proof_forms: bool = False) -> Proof:
    """Proves that
    `equivalence_of`('(~A`variable`[`formula`]', 'E`variable`[~`formula`]')`.

    Parameters:
        variable: variable name for the quantifications in the formula to be
            proven.
        formula: statement to be universally quantified, and whose negation is
            to be existentially quantified, in the formula to be proven.
        print_as_proof_forms: flag specifying whether the proof is to be printed
            in real time as it is being created.

    Returns:
        A valid proof of the above formula via
        `~predicates.prover.Prover.AXIOMS`.
    """
    from .prenex import equivalence_of
    assert is_variable(variable)
    prover = Prover({})
    proof1 = _prove_not_exists_not_implies_all(variable, formula)
    step1 = prover.add_proof(proof1.conclusion, proof1)
    proof2 = _prove_exists_not_implies_not_all(variable, formula)
    step2 = prover.add_proof(proof2.conclusion, proof2)
    conclusion = equivalence_of(
                    Formula('~', Formula('A', variable, formula)),
                    Formula('E', variable, Formula('~', formula)))
    step3 = prover.add_tautological_implication(conclusion, {step1, step2})
    return prover.qed()
    # Optional Task 11.4c


# prover.add_ug(quantified, nonquantified_line_number)
# prover.add_universal_instantiation(instantiation, line_number, term)
# prover.add_free_instantiation(instantiation, line_number, substitution_map)
# prover.add_flipped_equality(flipped, line_number)
# prover.add_substituted_equality(substituted, line_number, parametrized_term)
# prover.add_instantiated_assumption(instantiated, assumption, inst_map)
# prover.add_tautological_implication(implication, line_numbers)
# prover.add_existential_derivation(consequent, line_number1, line_number2) 1 = Ex[antecedent], 2 = '(antecedent(x)->consequent)'



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/soundness.py

"""Programmatic proof of the soundness of Propositional Logic."""

from typing import Tuple
from .syntax import *
from .semantics import *
from .proofs import *

def rule_nonsoundness_from_specialization_nonsoundness(
        general: InferenceRule, specialization: InferenceRule, model: Model) \
        -> Model:
    """Demonstrates the non-soundness of the given general inference rule given
    an example of the non-soundness of the given specialization of this rule.

    Parameters:
        general: inference rule to the soundness of which to find a
            counterexample.
        specialization: non-sound specialization of `general`.
        model: model in which `specialization` does not hold.

    Returns:
        A model in which `general` does not hold.
    """
    assert specialization.is_specialization_of(general)
    assert not evaluate_inference(specialization, model)

    map1 = general.specialization_map(specialization)
    g_model = dict()
    for key in map1.keys():
        g_model[key] = evaluate(map1[key], model)
    return g_model
    # Task 4.9

def nonsound_rule_of_nonsound_proof(proof: Proof, model: Model) -> \
        Tuple[InferenceRule, Model]:
    """Finds a non-sound inference rule used by the given valid proof of a
    non-sound inference rule, and demonstrates the non-soundness of the former
    rule.

    Parameters:
        proof: valid proof of a non-sound inference rule.
        model: model in which the inference rule proved by the given proof does
            not hold.

    Returns:
        A pair of a non-sound inference rule used in the given proof and a model
        in which this rule does not hold.
    """
    assert proof.is_valid()
    assert not evaluate_inference(proof.statement, model)
    for i in range(len(proof.lines)):
        if not proof.lines[i].is_assumption() and not evaluate(proof.lines[i].formula, model):
            suspect, number = proof.lines[i], i
            break
    return (suspect.rule, rule_nonsoundness_from_specialization_nonsoundness(suspect.rule, proof.rule_for_line(i), model))
    # Task 4.10



# mltp/predicates/__init__.py
from .syntax import (ForbiddenVariableError,
                     is_constant,
                     is_variable,
                     is_function,
                     Term,
                     is_equality,
                     is_relation,
                     is_unary,
                     is_binary,
                     is_quantifier,
                     Formula)
from .semantics import (T, 
                        Model)
from .functions import (function_name_to_relation_name,
                        relation_name_to_function_name,
                        replace_functions_with_relations_in_model,
                        replace_relations_with_functions_in_model,
                        _compile_term,
                        replace_functions_with_relations_in_formula,
                        replace_functions_with_relations_in_formulas,
                        translate_equality_to_SAME,
                        replace_equality_with_SAME_in_formulas,
                        add_SAME_as_equality_in_model,
                        make_equality_as_SAME_in_model)
from .proofs import (InstantiationMap,
                     Schema,
                     Proof,
                     I0_SCHEMA,
                     I1_SCHEMA,
                     D_SCHEMA,
                     I2_SCHEMA,
                     N_SCHEMA,
                     NI_SCHEMA,
                     NN_SCHEMA,
                     R_SCHEMA,
                     PROPOSITIONAL_AXIOMATIC_SYSTEM_SCHEMAS,
                     PROPOSITIONAL_AXIOM_TO_SCHEMA,
                     _axiom_specialization_map_to_schema_instantiation_map,
                     _prove_from_skeleton_proof,
                     prove_tautology,
                     is_tautology)
from .prover import Prover
from .deduction import (remove_assumption,
                        prove_by_way_of_contradiction)
from .prenex import (ADDITIONAL_QUANTIFICATION_AXIOMS,
                     is_quantifier_free,
                     is_in_prenex_normal_form,
                     equivalence_of,
                     has_uniquely_named_variables,
                     _uniquely_rename_quantified_variables_helper,
                     _uniquely_rename_quantified_variables,
                     _pull_out_quantifications_across_negation,
                     _pull_out_quantifications_from_left_across_binary_operator,
                     _pull_out_quantifications_from_right_across_binary_operator,
                     _pull_out_quantifications_across_binary_operator,
                     _to_prenex_normal_form_from_uniquely_named_variables,
                     to_prenex_normal_form)
from .completeness import (is_closed,
                           is_primitively_closed,
                           is_universally_closed,
                           is_existentially_closed,
                           find_unsatisfied_quantifier_free_sentence,
                           get_primitives,
                           get_constants,
                           model_or_inconsistency,
                           combine_contradictions,
                           eliminate_universal_instantiation_assumption,
                           universal_closure_step,
                           replace_constant,
                           eliminate_existential_witness_assumption,
                           existential_closure_step)
from .operators import frege
from .some_proofs import (prove_syllogism,
                          prove_syllogism_with_universal_instantiation,
                          prove_syllogism_all_all,
                          prove_syllogism_all_all_with_tautological_implication,
                          prove_syllogism_all_exists,
                          prove_syllogism_all_exists_with_existential_derivation,
                          prove_lovers,
                          prove_homework,
                          GROUP_AXIOMS,
                          prove_group_right_neutral,
                          prove_group_unique_zero,
                          FIELD_AXIOMS,
                          prove_field_zero_multiplication,
                          INDUCTION_AXIOM,
                          PEANO_AXIOMS,
                          prove_peano_left_neutral,
                          COMPREHENSION_AXIOM,
                          prove_russell_paradox,
                          _prove_not_exists_not_implies_all,
                          _prove_exists_not_implies_not_all,
                          prove_not_all_iff_exists_not)

#__all__ = ['semantics', 'completeness', 'deduction', 'functions', 'operators', 'prenex', 'proofs', 'prover', 'some_proofs', 'syntax']
__all__ = ['ForbiddenVariableError',
             'is_constant',
             'is_variable',
             'is_function',
             'Term',
             'is_equality',
             'is_relation',
             'is_unary',
             'is_binary',
             'is_quantifier',
             'Formula',
             'T',
             'Model',
             'function_name_to_relation_name',
             'relation_name_to_function_name',
             'replace_functions_with_relations_in_model',
             'replace_relations_with_functions_in_model',
             '_compile_term',
             'replace_functions_with_relations_in_formula',
             'replace_functions_with_relations_in_formulas',
             'translate_equality_to_SAME',
             'replace_equality_with_SAME_in_formulas',
             'add_SAME_as_equality_in_model',
             'make_equality_as_SAME_in_model',
             'InstantiationMap',
             'Schema',
             'Proof',
             'I0_SCHEMA',
             'I1_SCHEMA',
             'D_SCHEMA',
             'I2_SCHEMA',
             'N_SCHEMA',
             'NI_SCHEMA',
             'NN_SCHEMA',
             'R_SCHEMA',
             'PROPOSITIONAL_AXIOMATIC_SYSTEM_SCHEMAS',
             'PROPOSITIONAL_AXIOM_TO_SCHEMA',
             '_axiom_specialization_map_to_schema_instantiation_map',
             '_prove_from_skeleton_proof',
             'prove_tautology',
             'is_tautology',
             'Prover',
             'remove_assumption',
             'prove_by_way_of_contradiction',
             'ADDITIONAL_QUANTIFICATION_AXIOMS',
             'is_quantifier_free',
             'is_in_prenex_normal_form',
             'equivalence_of',
             'has_uniquely_named_variables',
             '_uniquely_rename_quantified_variables_helper',
             '_uniquely_rename_quantified_variables',
             '_pull_out_quantifications_across_negation',
             '_pull_out_quantifications_from_left_across_binary_operator',
             '_pull_out_quantifications_from_right_across_binary_operator',
             '_pull_out_quantifications_across_binary_operator',
             '_to_prenex_normal_form_from_uniquely_named_variables',
             'to_prenex_normal_form',
             'is_closed',
             'is_primitively_closed',
             'is_universally_closed',
             'is_existentially_closed',
             'find_unsatisfied_quantifier_free_sentence',
             'get_primitives',
             'get_constants',
             'model_or_inconsistency',
             'combine_contradictions',
             'eliminate_universal_instantiation_assumption',
             'universal_closure_step',
             'replace_constant',
             'eliminate_existential_witness_assumption',
             'existential_closure_step',
             'prove_syllogism',
             'prove_syllogism_with_universal_instantiation',
             'prove_syllogism_all_all',
             'prove_syllogism_all_all_with_tautological_implication',
             'prove_syllogism_all_exists',
             'prove_syllogism_all_exists_with_existential_derivation',
             'prove_lovers',
             'prove_homework',
             'GROUP_AXIOMS',
             'prove_group_right_neutral',
             'prove_group_unique_zero',
             'FIELD_AXIOMS',
             'prove_field_zero_multiplication',
             'INDUCTION_AXIOM',
             'PEANO_AXIOMS',
             'prove_peano_left_neutral',
             'COMPREHENSION_AXIOM',
             'prove_russell_paradox',
             '_prove_not_exists_not_implies_all',
             '_prove_exists_not_implies_not_all',
             'prove_not_all_iff_exists_not']


# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/operators.py

"""Syntactic conversion of propositional formulas to use only specific sets of
operators."""
from .syntax import *
from .semantics import *

def to_not_and_or(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '~', '&', and '|'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '~', '&', and
        '|'.
    """
    sub_map = { 'T': Formula.parse('(p|~p)'),
                'F': Formula.parse('(p&~p)'),
                '->': Formula.parse('(~p|q)'),
                '+': Formula.parse('((p|q)&~(p&q))'),
                '-&': Formula.parse('~(p&q)'),
                '-|': Formula.parse('~(p|q)'),
                '<->': Formula.parse('((~p|q)&(~q|p))')}
    return Formula.substitute_operators(formula, sub_map)
    # Task 3.5

def to_not_and(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '~' and '&'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '~' and '&'.
    """
    sub_map = { 'T': Formula.parse('~(p&~p)'),
                'F': Formula.parse('(p&~p)'),
                '|': Formula.parse('~(~p&~q)'),
                '->': Formula.parse('~(p&~q)'),
                '+': Formula.parse('(~(p&q)&~(~p&~q))'),
                '-&': Formula.parse('~(p&q)'),
                '-|': Formula.parse('(~p&~q)'),
                '<->': Formula.parse('(~(p&~q)&~(q&~p))')}
    return Formula.substitute_operators(formula, sub_map)
    # Task 3.6a

def to_nand(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '-&'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '-&'.
    """
    sub_map = { 'T': Formula.parse('(p-&(p-&p))'),
                'F': Formula.parse('((p-&(p-&p))-&(p-&(p-&p)))'),
                '~': Formula.parse('(p-&p)'),
                '&': Formula.parse('((p-&q)-&(p-&q))'),
                '|': Formula.parse('((p-&p)-&(q-&q))'),
                '->': Formula.parse('(p-&(q-&q))'),
                '+': Formula.parse('((p-&(p-&q))-&(q-&(p-&q)))'),
                '-|': Formula.parse('(((p-&p)-&(q-&q))-&((p-&p)-&(q-&q)))'),
                '<->': Formula.parse('(((p-&p)-&(q-&q))-&(p-&q))')}
    return Formula.substitute_operators(formula, sub_map)
    # Task 3.6b

def to_implies_not(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '->' and '~'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '->' and '~'.
    """
    sub_map = { 'T': Formula.parse('(p->p)'),
                'F': Formula.parse('~(p->p)'),
                '&': Formula.parse('~(p->~q)'),
                '|': Formula.parse('(~p->q)'),
                '+': Formula.parse('((p->q)->~(q->p))'),
                '-|': Formula.parse('~(~p->q)'),
                '-&': Formula.parse('(p->~q)'),
                '<->': Formula.parse('~((p->q)->~(q->p))')}
    return Formula.substitute_operators(formula, sub_map)
    # Task 3.6c

def to_implies_false(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '->' and 'F'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '->' and 'F'.
    """
    sub_map = { 'T': Formula.parse('(F->p)'),
                '~': Formula.parse('(p->F)'),
                '&': Formula.parse('((p->(q->F))->F)'),
                '|': Formula.parse('((p->q)->q)'),
                '+': Formula.parse('((p->q)->((q->p)->F))'),
                '-|': Formula.parse('(((p->q)->q)->F)'),
                '-&': Formula.parse('((p->q)->(p->F))'),
                '<->': Formula.parse('(((p->q)->((q->p)->F))->F)')}
    return Formula.substitute_operators(formula, sub_map)
    # Task 3.6d

def to_not_or(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '~' and '|'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '~' and '|'.
    """
    sub_map = { 'T': Formula.parse('(p|~p)'),
                'F': Formula.parse('~(p|~p)'),
                '&': Formula.parse('~(~p|~q)'),
                '->': Formula.parse('~(~p|q)'),
                '+': Formula.parse('(~(~p|q)|~(~q|p))'),
                '-&': Formula.parse('(~p|~q)'),
                '-|': Formula.parse('~(p|q)'),
                '<->': Formula.parse('~(~(~p|q)|~(~q|p))')}
    return Formula.substitute_operators(formula, sub_map)
    # Personal task

def to_nor(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '-|'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '-|'.
    """
    sub_map = { 'T': Formula.parse('((p-|(p-|p))-|(p-|(p-|p)))'),
                'F': Formula.parse('(p-|(p-|p))'),
                '~': Formula.parse('(p-|p)'),
                '&': Formula.parse('((p-|p)-|(q-|q))'),
                '|': Formula.parse('((p-|q)-|(p-|q))'),
                '->': Formula.parse('(((p-|p)-|q)-|((p-|p)-|q))'),
                '+': Formula.parse('((p-|q)-|((p-|p)-|(q-|q)))'),
                '-&': Formula.parse('(((p-|p)-|(q-|q))-|((p-|p)-|(q-|q)))'),
                '<->': Formula.parse('((p-|(p-|q))-|(q-|(p-|q)))')}
    return Formula.substitute_operators(formula, sub_map)
    # Personal task

#(to_not_and_or, to_not_and, to_nand, to_implies_not, to_implies_false, to_not_or, to_nor)



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/some_proofs.py

"""Some proofs in Propositional Logic."""
from .syntax import *
from .proofs import *
from .axiomatic_systems import *
from .deduction import *

# Some inference rules that only use conjunction.

#: Conjunction introduction inference rule
A_RULE = InferenceRule([Formula.parse('x'), Formula.parse('y')],
                       Formula.parse('(x&y)'))
#: Conjunction elimination (right) inference rule
AE1_RULE = InferenceRule([Formula.parse('(x&y)')],Formula.parse('y'))
#: Conjunction elimination (left) inference rule
AE2_RULE = InferenceRule([Formula.parse('(x&y)')],Formula.parse('x'))

def prove_and_commutativity() -> Proof:
    """Proves ``'(q&p)'`` from ``'(p&q)'`` via `A_RULE`, `AE1_RULE`, and
    `AE2_RULE`.

    Returns:
        A valid proof of ``'(q&p)'`` from the single assumption ``'(p&q)'`` via
        the inference rules `A_RULE`, `AE1_RULE`, and `AE2_RULE`.
    """
    statement = InferenceRule([Formula.parse('(p&q)')], Formula.parse('(q&p)'))
    rules = {A_RULE, AE1_RULE, AE2_RULE}
    lines = [Proof.Line(Formula.parse('(p&q)')),
             Proof.Line(Formula.parse('q'), AE1_RULE, [0]),
             Proof.Line(Formula.parse('p'), AE2_RULE, [0]),
             Proof.Line(Formula.parse('(q&p)'), A_RULE, [1,2])]

    proof1 = Proof(statement, rules, lines)
    return proof1
    # Task 4.7

def prove_I0() -> Proof:
    """Proves `~propositions.axiomatic_systems.I0` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I1`,
    and `~propositions.axiomatic_systems.D`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.I0` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I1`, and
        `~propositions.axiomatic_systems.D`.
    """

    return Proof(
    I0,
    [MP, D, I1],
	[Proof.Line(Formula.parse('((p->((p->p)->p))->((p->(p->p))->(p->p)))'), D, []),
		Proof.Line(Formula.parse('(p->((p->p)->p))'), I1, []),
		Proof.Line(Formula.parse('((p->(p->p))->(p->p))'), MP, [1, 0]),
		Proof.Line(Formula.parse('(p->(p->p))'), I1, []),
		Proof.Line(Formula.parse('(p->p)'), MP, [3, 2])])
    # Task 4.8

#: Hypothetical syllogism
HS = InferenceRule([Formula.parse('(p->q)'), Formula.parse('(q->r)')],
                   Formula.parse('(p->r)'))

def prove_hypothetical_syllogism() -> Proof:
    """Proves `HS` via `~propositions.axiomatic_systems.MP`,
    `~propositions.axiomatic_systems.I0`, `~propositions.axiomatic_systems.I1`,
    and `~propositions.axiomatic_systems.D`.

    Returns:
        A valid proof of `HS` via the inference rules
        `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`, and
        `~propositions.axiomatic_systems.D`.
    """
    return Proof(
	HS,
	[MP, I0, I1, D],
	[Proof.Line(Formula.parse('(p->q)')),
		Proof.Line(Formula.parse('(q->r)')),
		Proof.Line(Formula.parse('((q->r)->(p->(q->r)))'), I1, []),
		Proof.Line(Formula.parse('(p->(q->r))'), MP, [1, 2]),
		Proof.Line(Formula.parse('((p->(q->r))->((p->q)->(p->r)))'), D, []),
		Proof.Line(Formula.parse('((p->q)->(p->r))'), MP, [3, 4]),
		Proof.Line(Formula.parse('(p->r)'), MP, [0, 5])])
    # Task 5.5

def prove_I2() -> Proof:
    """Proves `~propositions.axiomatic_systems.I2` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    and `~propositions.axiomatic_systems.N`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.I2` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    # Optional Task 6.7a

#: Double-negation elimination
_NNE = InferenceRule([], Formula.parse('(~~p->p)'))

def _prove_NNE() -> Proof:
    """Proves `_NNE` via `~propositions.axiomatic_systems.MP`,
    `~propositions.axiomatic_systems.I0`, `~propositions.axiomatic_systems.I1`,
    `~propositions.axiomatic_systems.D`, and
    `~propositions.axiomatic_systems.N`.

    Returns:
        A valid proof of `_NNE` via the inference rules
        `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    # Optional Task 6.7b

def prove_NN() -> Proof:
    """Proves `~propositions.axiomatic_systems.NN` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    and `~propositions.axiomatic_systems.N`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.NN` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    # Optional Task 6.7c

#: Contraposition
_CP = InferenceRule([], Formula.parse('((p->q)->(~q->~p))'))

def _prove_CP() -> Proof:
    """Proves `_CP` via `~propositions.axiomatic_systems.MP`,
    `~propositions.axiomatic_systems.I0`, `~propositions.axiomatic_systems.I1`,
    `~propositions.axiomatic_systems.D`, and
    `~propositions.axiomatic_systems.N`.

    Returns:
        A valid proof of `_CP` via the inference rules
        `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    # Optional Task 6.7d

def prove_NI() -> Proof:
    """Proves `~propositions.axiomatic_systems.NI` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    and `~propositions.axiomatic_systems.N`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.NI` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    # Optional Task 6.7e

#: Consequentia mirabilis
_CM = InferenceRule([Formula.parse('(~p->p)')], Formula.parse('p'))

def _prove_CM() -> Proof:
    """Proves `_CM` via `~propositions.axiomatic_systems.MP`,
    `~propositions.axiomatic_systems.I0`, `~propositions.axiomatic_systems.I1`,
    `~propositions.axiomatic_systems.D`, and
    `~propositions.axiomatic_systems.N`.

    Returns:
        A valid proof of `_CM` via the inference rules
        `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    # Optional Task 6.7f

def prove_R() -> Proof:
    """Proves `~propositions.axiomatic_systems.R` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    and `~propositions.axiomatic_systems.N`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.R` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N`.
    """
    # Optional Task 6.7g

def prove_N() -> Proof:
    """Proves `~propositions.axiomatic_systems.N` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    and `~propositions.axiomatic_systems.N_ALTERNATIVE`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.N` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.N_ALTERNATIVE`.
    """
    # Optional Task 6.8

def prove_NA1() -> Proof:
    """Proves `~propositions.axiomatic_systems.NA1` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    `~propositions.axiomatic_systems.N`, and
    `~propositions.axiomatic_systems.AE1`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.NA1` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.AE1`.
    """
    # Optional Task 6.9a

def prove_NA2() -> Proof:
    """Proves `~propositions.axiomatic_systems.NA2` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    `~propositions.axiomatic_systems.N`, and
    `~propositions.axiomatic_systems.AE2`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.NA2` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.AE2`.
    """
    # Optional Task 6.9b

def prove_NO() -> Proof:
    """Proves `~propositions.axiomatic_systems.NO` via
    `~propositions.axiomatic_systems.MP`, `~propositions.axiomatic_systems.I0`,
    `~propositions.axiomatic_systems.I1`, `~propositions.axiomatic_systems.D`,
    `~propositions.axiomatic_systems.N`, and
    `~propositions.axiomatic_systems.OE`.

    Returns:
        A valid proof of `~propositions.axiomatic_systems.NO` via the inference
        rules `~propositions.axiomatic_systems.MP`,
        `~propositions.axiomatic_systems.I0`,
        `~propositions.axiomatic_systems.I1`,
        `~propositions.axiomatic_systems.D`, and
        `~propositions.axiomatic_systems.OE`.
    """
    # Optional Task 6.9c




# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/reductions.py

"""Reduction between computational search problems."""

from __future__ import annotations
from typing import AbstractSet, Mapping, Tuple, Union
from .syntax import *
from .semantics import *
from .operators import *
#import itertools

#: A graph on a vertex set of the form `(1, ..., n_vertices)`,
#: represented by the number of vertices `n_vertices` and a set of edges over
#: the vertices.
Graph = Tuple[int, AbstractSet[Tuple[int, int]]]

def is_graph(graph: Graph) -> bool:
    """Checks if the given data structure is a valid representation of a graph.

    Parameters:
        graph: data structure to check.

    Returns:
        ``True`` if the given data structure is a valid representation of a
        graph, ``False`` otherwise.
    """
    (n_vertices, edges) = graph
    for edge in edges:
        for vertex in edge:
            if not 1 <= vertex <= n_vertices:
                return False
        if edge[0] == edge[1]:
            return False
    return True

def is_valid_3coloring(graph: Graph, coloring: Mapping[int, int]) -> bool:
    """Checks whether the given coloring is a valid coloring of the given graph
    by the colors 1, 2, and 3.

    Parameters:
        graph: graph to check.
        coloring: mapping from the vertices of the given graph to colors,
            to check.

    Returns:
        ``True`` if the given coloring is a valid coloring of the given graph by
        the colors 1, 2, and 3; ``False`` otherwise.
    """
    assert is_graph(graph)
    (n_vertices, edges) = graph
    for vertex in range(1, n_vertices + 1):
        if vertex not in coloring.keys() or coloring[vertex] not in {1, 2, 3}:
            return False
    for edge in edges:
        if coloring[edge[0]] == coloring[edge[1]]:
            return False
    return True

def graph3coloring_to_formula(graph: Graph) -> Formula:
    """Efficiently reduces the 3-coloring problem of the given graph into a
    satisfiability problem.

    Parameters:
        graph: graph whose 3-coloring problem to reduce.

    Returns:
        A propositional formula that is satisfiable if and only if the given
        graph is 3-colorable.
    """
    assert is_graph(graph)
    (n_vertices, edges) = graph
    colors = ('1','2','3')
    formula = None

    for node in range(1,n_vertices+1):
        variables = ['x'+str(node)+color for color in colors]
        nodeformula = Formula(variables[0])

        for variable in variables[1:]:
            nodeformula = Formula('|',nodeformula,Formula(variable))

        for i,j in combinations(variables,2):
            tempformula = Formula('|',Formula('~',Formula(i)),Formula('~',Formula(j)))
            nodeformula = Formula('&',nodeformula,tempformula)

        if not formula:
            formula = nodeformula
        else:
            formula = Formula('&',formula,nodeformula)

    for node_0,node_1 in edges:
        variables = zip(('x'+str(node_0)+color for color in colors),('x'+str(node_1)+color for color in colors))
        for source,target in variables:
            formula = Formula('&', formula, Formula('|', Formula('~', Formula(source)), Formula('~', Formula(target))))
    #print(formula)
    return formula
    # Optional Task 2.10a

def assignment_to_3coloring(graph: Graph, assignment: Model) -> \
        Mapping[int, int]:
    """Efficiently transforms an assignment to the formula corresponding to the
    3-coloring problem of the given graph, to a 3-coloring of the given graph so
    that the 3-coloring is valid if and only if the given assignment is
    satisfying.

    Parameters:
        graph: graph to produce a 3-coloring for.
        assignment: assignment to the variable names of the formula returned by
            `graph3coloring_to_formula(graph)`.

    Returns:
        A 3-coloring of the given graph by the colors 1, 2, and 3 that is valid
        if and only if the given assignment satisfies the formula
        `graph3coloring_to_formula(graph)`.
    """
    assert is_graph(graph)
    formula = graph3coloring_to_formula(graph)
    assert evaluate(formula, assignment)

    mapping = dict()
    for i in formula.variables():
        if assignment[i]:
            mapping[int(i[1:-1])] = int(i[-1])
    return mapping
    # Optional Task 2.10b

def tricolor_graph(graph: Graph) -> Union[Mapping[int, int], None]:
    """Computes a 3-coloring of the given graph.

    Parameters:
        graph: graph to 3-color.

    Returns:
        An arbitrary 3-coloring of the given graph if it is 3-colorable,
        ``None`` otherwise.
    """
    assert is_graph(graph)
    formula = graph3coloring_to_formula(graph)
    for assignment in all_models(list(formula.variables())):
        if evaluate(formula, assignment):
            return assignment_to_3coloring(graph, assignment)
    return None

# graph1 = (7, {(1,6),(2,6),(3,6),(4,6),(5,6),(6,7),(1,4),(2,4),(2,3),(3,5),(4,6),(4,5),(1,7),(2,7),(3,7),(4,7)})

# graph = (8, {(1,6),(2,6),(3,6),(4,6),(5,6),(6,7),(6,8),(1,4),(1,7),(2,4),(2,3),(3,5),(4,6),(4,5),(4,7),(5,7),(5,8),(7,8)})
# print(tricolor_graph(graph1))

# graph = (3, {(1,2) ,(1,3), (2,3)})
# assignment = {'x11':True, 'x12':False, 'x13':False, 'x21':True, 'x22':False, 'x23':False, 'x31':False, 'x32':True, 'x33':False, 'x41':False, 'x42':True, 'x43':False, 'x51':True, 'x52':False, 'x53':False, 'x61':False, 'x62':False, 'x63':True, 'x71':False, 'x72':False, 'x73':True, 'x81':False, 'x82':True, 'x83':False}
# formula = graph3coloring_to_formula(graph)
# print(formula,'\n')
# print(tricolor_graph(graph))


# one = list(range(1,3))
# two = one[1:]
# two.append(one[0])
# three = set(zip(one, two))
# graph = (len(three), three)
# formula = graph3coloring_to_formula(graph)
# print(formula)
# print('\n\n',to_not_or(formula))
